"""Classes for defining neural network weight optimization problems."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause


import numpy as np
from abc import abstractmethod
from sklearn.preprocessing import LabelBinarizer
from mlrose_ky.algorithms.decay import GeometricDecay
from mlrose_ky.algorithms.rhc import random_hill_climb
from mlrose_ky.algorithms.sa import simulated_annealing
from mlrose_ky.algorithms.ga import genetic_alg

from mlrose_ky.neural._nn_base import _NNBase
from mlrose_ky.neural.activation import identity, relu, sigmoid, tanh
from mlrose_ky.neural.utils.weights import gradient_descent_original


class _NNCore(_NNBase):
    """Core class for neural networks.

    Warning: This class should not be used directly.
    Use derived classes instead.
    """

    @abstractmethod
    def __init__(
        self,
        hidden_nodes=None,
        activation="relu",
        algorithm="random_hill_climb",
        max_iters=100,
        bias=True,
        is_classifier=True,
        learning_rate=0.1,
        early_stopping=False,
        clip_max=1e10,
        restarts=0,
        schedule=GeometricDecay(),
        pop_size=200,
        mutation_prob=0.1,
        max_attempts=10,
        random_state=None,
        curve=False,
    ):

        super().__init__()
        if hidden_nodes is None:
            self.hidden_nodes = []
        else:
            self.hidden_nodes = hidden_nodes

        self.activation_dict = {"identity": identity, "relu": relu, "sigmoid": sigmoid, "tanh": tanh}
        self.activation = activation
        self.algorithm = algorithm
        self.max_iters = max_iters
        self.bias = bias
        self.is_classifier = is_classifier
        self.learning_rate = learning_rate
        self.early_stopping = early_stopping
        self.clip_max = clip_max
        self.restarts = restarts
        self.schedule = schedule
        self.pop_size = pop_size
        self.mutation_prob = mutation_prob
        self.max_attempts = max_attempts
        self.random_state = random_state
        self.curve = curve

        self.node_list = []
        self.fitted_weights = []
        self.loss = np.inf
        self.output_activation = None
        self.predicted_probs = []
        self.fitness_curve = []

    def _validate(self):
        if (not isinstance(self.max_iters, int) and self.max_iters != np.inf and not self.max_iters.is_integer()) or self.max_iters < 0:
            raise Exception("""max_iters must be a positive integer.""")

        if not isinstance(self.bias, bool):
            raise Exception("""bias must be True or False.""")

        if not isinstance(self.is_classifier, bool):
            raise Exception("""is_classifier must be True or False.""")

        if self.learning_rate <= 0:
            raise Exception("""learning_rate must be greater than 0.""")

        if not isinstance(self.early_stopping, bool):
            raise Exception("""early_stopping must be True or False.""")

        if self.clip_max <= 0:
            raise Exception("""clip_max must be greater than 0.""")

        if (not isinstance(self.max_attempts, int) and not self.max_attempts.is_integer()) or self.max_attempts < 0:
            raise Exception("""max_attempts must be a positive integer.""")

        if self.pop_size < 0:
            raise Exception("""pop_size must be a positive integer.""")
        elif not isinstance(self.pop_size, int):
            if self.pop_size.is_integer():
                self.pop_size = int(self.pop_size)
            else:
                raise Exception("""pop_size must be a positive integer.""")

        if self.mutation_prob < 0 or self.mutation_prob > 1:
            raise Exception("""mutation_prob must be between 0 and 1.""")

        if self.activation is None or self.activation not in self.activation_dict.keys():
            raise Exception(
                """Activation function must be one of: 'identity',
                    'relu', 'sigmoid' or 'tanh'."""
            )

        if self.algorithm not in ["random_hill_climb", "simulated_annealing", "genetic_alg", "gradient_descent"]:
            raise Exception("""Algorithm must be one of: 'random_hill_climb', 'simulated_annealing', 'genetic_alg', 'gradient_descent'.""")

    def _validate_input(self, y):
        """
        Add classes_ attribute based on classes present in y.
        """
        # Copied from
        # https://github.com/scikit-learn/scikit-learn/blob/5c4aa5d0d90ba66247d675d4c3fc2fdfba3c39ff/sklearn/neural_network/_multilayer_perceptron.py
        # Note: Required for scikit-learn 1.3+. Doesn't cause issues for lower versions.
        # No workaround found for multi-class labels, still doesn't work with f1 score.
        if not hasattr(self, "classes_"):
            self._label_binarizer = LabelBinarizer()
            self._label_binarizer.fit(y)
            self.classes_ = self._label_binarizer.classes_

    def fit(self, X, y=None, init_weights=None):
        """Fit neural network to data.

        Parameters
        ----------
        X: np.ndarray
            Numpy array containing feature dataset with each row
            representing a single observation.

        y: np.ndarray
            Numpy array containing data labels. Length must be same as
            length of X.

        init_weights: np.ndarray, default: None
            Numpy array containing starting weights for algorithm.
            If :code:`None`, then a random state is used.
        """
        self._validate()
        self._validate_input(y)

        X, y = self._format_x_y_data(X, y)

        node_list = self._build_node_list(X, y, self.hidden_nodes, self.bias)
        num_nodes = self._calculate_state_size(node_list)

        if init_weights is not None and len(init_weights) != num_nodes:
            raise Exception("""init_weights must be None or have length %d""" % (num_nodes,))

        # Set random seed
        if isinstance(self.random_state, int) and self.random_state > 0:
            np.random.seed(self.random_state)

        fitness, problem = self._build_problem_and_fitness_function(
            X, y, node_list, self.activation_dict[self.activation], self.learning_rate, self.bias, self.clip_max, self.is_classifier
        )

        if self.algorithm == "random_hill_climb":
            fitness_curve, fitted_weights, loss = self.__run_with_rhc(init_weights, num_nodes, problem)

        elif self.algorithm == "simulated_annealing":
            fitness_curve, fitted_weights, loss = self._run_with_sa(init_weights, num_nodes, problem)
        elif self.algorithm == "genetic_alg":
            fitness_curve, fitted_weights, loss = self._run_with_ga(problem)
        else:  # Gradient descent case
            fitness_curve, fitted_weights, loss = self._run_with_gd(init_weights, num_nodes, problem)

        # Save fitted weights and node list
        self.node_list = node_list
        self.fitted_weights = fitted_weights
        self.loss = loss
        self.output_activation = fitness.get_output_activation()

        if self.curve:
            self.fitness_curve = fitness_curve

        return self

    def _run_with_gd(self, init_weights, num_nodes, problem):
        if init_weights is None:
            init_weights = np.random.uniform(-1, 1, num_nodes)

        fitted_weights, loss, fitness_curve = gradient_descent_original(
            problem,
            max_attempts=self.max_attempts if self.early_stopping else self.max_iters,
            max_iters=self.max_iters,
            curve=self.curve,
            init_state=init_weights,
        )

        return ([] if fitness_curve is None else fitness_curve), fitted_weights, loss

    def _run_with_ga(self, problem):
        fitness_curve = []
        if self.curve:
            fitted_weights, loss, fitness_curve = genetic_alg(
                problem,
                pop_size=self.pop_size,
                mutation_prob=self.mutation_prob,
                max_attempts=self.max_attempts if self.early_stopping else self.max_iters,
                max_iters=self.max_iters,
                curve=self.curve,
            )
        else:
            fitted_weights, loss, _ = genetic_alg(
                problem,
                pop_size=self.pop_size,
                mutation_prob=self.mutation_prob,
                max_attempts=self.max_attempts if self.early_stopping else self.max_iters,
                max_iters=self.max_iters,
                curve=self.curve,
            )
        return fitness_curve, fitted_weights, loss

    def _run_with_sa(self, init_weights, num_nodes, problem):
        fitness_curve = []
        if init_weights is None:
            init_weights = np.random.uniform(-1, 1, num_nodes)
        if self.curve:
            fitted_weights, loss, fitness_curve = simulated_annealing(
                problem,
                schedule=self.schedule,
                max_attempts=self.max_attempts if self.early_stopping else self.max_iters,
                max_iters=self.max_iters,
                init_state=init_weights,
                curve=self.curve,
            )
        else:
            fitted_weights, loss, _ = simulated_annealing(
                problem,
                schedule=self.schedule,
                max_attempts=self.max_attempts if self.early_stopping else self.max_iters,
                max_iters=self.max_iters,
                init_state=init_weights,
                curve=self.curve,
            )
        return fitness_curve, fitted_weights, loss

    def __run_with_rhc(self, init_weights, num_nodes, problem):
        fitness_curve = []
        fitted_weights = []
        loss = np.inf

        # Can't use restart feature of random_hill_climb function, since
        # want to keep initial weights in the range -1 to 1.
        for _ in range(self.restarts + 1):
            restart_weights = np.random.uniform(-1, 1, num_nodes) if init_weights is None else init_weights

            current_weights, current_loss, fitness_curve = random_hill_climb(
                problem,
                max_attempts=(self.max_attempts if self.early_stopping else self.max_iters),
                max_iters=self.max_iters,
                init_state=restart_weights,
                curve=self.curve,
            )

            if not self.curve:
                fitness_curve = []

            if current_loss < loss:
                fitted_weights = current_weights
                loss = current_loss

        return fitness_curve, fitted_weights, loss

    def predict(self, X):
        """Use model to predict data labels for given feature array.

        Parameters
        ----------
        X: np.ndarray
            Numpy array containing feature dataset with each row
            representing a single observation.

        Returns
        -------
        y_pred: np.ndarray
            Numpy array containing predicted data labels.
        """
        if not np.shape(X)[1] == (self.node_list[0] - self.bias):
            raise Exception("""The number of columns in X must equal %d""" % ((self.node_list[0] - self.bias),))

        y_pred, pp = self._predict(
            X=X,
            fitted_weights=self.fitted_weights,
            node_list=self.node_list,
            input_activation=self.activation_dict[self.activation],
            output_activation=self.output_activation,
            bias=self.bias,
            is_classifier=self.is_classifier,
        )
        self.predicted_probs = pp
        return y_pred



"""Classes for defining neural network weight optimization problems."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

from .network_weights import NetworkWeights



"""Class defining the Travelling Salesperson fitness function for use with optimization algorithms."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

import numpy as np
from typing import Callable


class TravellingSalesperson:
    """
    Fitness function for the Travelling Salesperson optimization problem.

    Evaluates the fitness of a tour of n nodes, represented by state vector x, giving the order in which
    the nodes are visited, as the total distance travelled on the tour (including the distance travelled
    between the final node in the state vector and the first node in the state vector during the return
    leg of the tour). Each node must be visited exactly once for a tour to be considered valid.

    Parameters
    ----------
    coords : list[tuple[float, float]] | None, optional
        Ordered list of the (x, y) coordinates of all nodes (where element i gives the coordinates of node i).
        This assumes that travel between all pairs of nodes is possible. If this is not the case, then use
        distances instead.

    distances : list[tuple[int, int, float]] | None, optional
        List giving the distances, d, between all pairs of nodes, u and v, for which travel is possible, with each
        list item in the form (u, v, d). Order of the nodes does not matter, so (u, v, d) and (v, u, d) are
        considered to be the same. If a pair is missing from the list, it is assumed that travel between the two nodes
        is not possible. This argument is ignored if coords is not None.

    Examples
    --------
    >>> import mlrose_ky
    >>> import numpy as np
    >>> coords = [(0, 0), (3, 0), (3, 2), (2, 4), (1, 3)]
    >>> dists = [(0, 1, 3), (0, 2, 5), (0, 3, 1), (0, 4, 7), (1, 3, 6), (4, 1, 9), (2, 3, 8), (2, 4, 2), (3, 2, 8), (3, 4, 4)]
    >>> fitness_coords = mlrose_ky.TravellingSalesperson(coords=coords)
    >>> state = np.array([0, 1, 4, 3, 2])
    >>> fitness_coords.evaluate(state)
    13.861384090800865

    >>> fitness_dists = mlrose_ky.TravellingSalesperson(distances=dists)
    >>> fitness_dists.evaluate(state)
    29.0

    Note
    ----
    1. The TravellingSalesperson fitness function is suitable for use in travelling salesperson (tsp) optimization
       problems only.
    2. It is necessary to specify at least one of coords and distances in initializing a TravellingSalesperson
       fitness function object.
    """

    def __init__(self, coords: list[tuple] = None, distances: list[tuple] = None) -> None:
        # Ensure that at least one of coords or distances is provided
        if coords is None and distances is None:
            raise ValueError("At least one of coords and distances must be specified.")

        # Initialize class variables
        self.prob_type: str = "tsp"
        self.coords: list = coords
        self.distances: list = distances
        self.is_coords: bool = coords is not None

        # Determine which fitness calculation method to use
        self.calculate_fitness: Callable = self._calculate_fitness_by_coords if self.is_coords else self._calculate_fitness_by_distance

        if self.is_coords:
            # Precompute the coordinates array for faster access
            self.coords_array: np.ndarray = np.array(coords)
        else:
            # Remove duplicates and sort distances
            self.distances = list({tuple(sorted((u, v)) + [d]) for u, v, d in distances})

            # Unpack node lists and distances
            node1_list, node2_list, self.dist_list = zip(*self.distances)

            # Validation checks on distances
            if min(self.dist_list) <= 0:
                raise ValueError("The distance between each pair of nodes must be greater than 0.")
            if min(node1_list + node2_list) < 0:
                raise ValueError("The minimum node value must be 0.")
            if not max(node1_list + node2_list) == (len(set(node1_list + node2_list)) - 1):
                raise ValueError("All nodes must appear at least once in distances.")

            # Create a distance matrix for quick lookup of distances between nodes
            num_nodes = max(max(node1_list), max(node2_list)) + 1
            self.distance_matrix = np.full((num_nodes, num_nodes), np.inf)
            for u, v, d in self.distances:
                self.distance_matrix[u, v] = d
                self.distance_matrix[v, u] = d

    def evaluate(self, state: np.ndarray) -> float:
        """
        Evaluate the fitness of a state vector.

        Parameters
        ----------
        state : np.ndarray
            State array for evaluation. Each integer between 0 and (len(state) - 1), inclusive must appear exactly once in the array.

        Returns
        -------
        fitness : float
            Value of fitness function. Returns np.inf if travel between two consecutive nodes on the tour is not possible.
        """
        # Validation checks on the state array
        if self.is_coords and len(state) != len(self.coords):
            raise ValueError("state must have the same length as coords.")
        if not len(state) == len(set(state)):
            raise ValueError("Each node must appear exactly once in state.")
        if min(state) < 0:
            raise ValueError("All elements of state must be non-negative integers.")
        if max(state) >= len(state):
            raise ValueError("All elements of state must be less than len(state).")

        # Calculate and return the fitness of the state
        return float(self.calculate_fitness(state))

    def _calculate_fitness_by_coords(self, state: np.ndarray) -> float:
        """
        Calculate fitness based on coordinates.

        Parameters
        ----------
        state : np.ndarray
            State array for evaluation.

        Returns
        -------
        fitness : float
            Calculated fitness value.
        """
        # Map state indices to coordinates
        nodes = self.coords_array[state]

        # Calculate total journey distance using Euclidean distance
        journey = np.linalg.norm(nodes[1:] - nodes[:-1], axis=1).sum()
        journey += np.linalg.norm(nodes[0] - nodes[-1])

        return journey

    def _calculate_fitness_by_distance(self, state: np.ndarray) -> float:
        """
        Calculate fitness based on distances.

        Parameters
        ----------
        state : np.ndarray
            State array for evaluation.

        Returns
        -------
        fitness : float
            Calculated fitness value. Returns np.inf if any segment of the tour is not possible.
        """
        total_distance = 0.0
        num_nodes = len(state)

        # Iterate over each node in the state
        for i in range(num_nodes):
            start = state[i]
            end = state[(i + 1) % num_nodes]
            distance = self.distance_matrix[start, end]

            # Check if the segment is possible
            if np.isinf(distance):
                return np.inf

            total_distance += distance

        return total_distance

    def get_problem_type(self) -> str:
        """
        Return the problem type.

        Returns
        -------
        prob_type : str
            Specifies problem type as 'tsp'.
        """
        return self.prob_type



"""Unit tests for opt_probs/"""

# Author: Genevieve Hayes (modified by Kyle Nakamura)
# License: BSD 3 clause

import numpy as np

try:
    import mlrose_ky
except ImportError:
    import sys

    sys.path.append("..")
    import mlrose_ky

from mlrose_ky import OneMax, DiscreteOpt, ContinuousOpt, TSPOpt, OnePointCrossover
from mlrose_ky.neural import NetworkWeights
from mlrose_ky.neural.activation import identity
from mlrose_ky.opt_probs.opt_prob import OptProb


class TestOptProb:
    """Tests for _OptProb class."""

    def test_set_state_max(self):
        """Test set_state method for a maximization problem"""
        problem = OptProb(5, OneMax())
        x = np.array([0, 1, 2, 3, 4])
        problem.set_state(x)
        assert np.array_equal(problem.get_state(), x) and problem.get_fitness() == 10

    def test_set_state_min(self):
        """Test set_state method for a minimization problem"""
        problem = OptProb(5, OneMax(), maximize=False)
        x = np.array([0, 1, 2, 3, 4])
        problem.set_state(x)
        assert np.array_equal(problem.get_state(), x) and problem.get_fitness() == -10

    def test_set_population_max(self):
        """Test set_population method for a maximization problem"""
        problem = OptProb(5, OneMax())
        pop = np.array(
            [
                [0, 0, 0, 0, 1],
                [1, 0, 1, 0, 1],
                [1, 1, 1, 1, 0],
                [1, 0, 0, 0, 1],
                [100, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [1, 1, 1, 1, 1],
                [0, 0, 0, 0, -50],
            ]
        )
        pop_fit = np.array([1, 3, 4, 2, 100, 0, 5, -50])
        problem.set_population(pop)
        assert np.array_equal(problem.get_population(), pop) and np.array_equal(problem.get_pop_fitness(), pop_fit)

    def test_set_population_min(self):
        """Test set_population method for a minimization problem"""
        problem = OptProb(5, OneMax(), maximize=False)
        pop = np.array(
            [
                [0, 0, 0, 0, 1],
                [1, 0, 1, 0, 1],
                [1, 1, 1, 1, 0],
                [1, 0, 0, 0, 1],
                [100, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [1, 1, 1, 1, 1],
                [0, 0, 0, 0, -50],
            ]
        )
        pop_fit = -np.array([1, 3, 4, 2, 100, 0, 5, -50])
        problem.set_population(pop)
        assert np.array_equal(problem.get_population(), pop) and np.array_equal(problem.get_pop_fitness(), pop_fit)

    def test_best_child_max(self):
        """Test best_child method for a maximization problem"""
        problem = OptProb(5, OneMax())
        pop = np.array(
            [
                [0, 0, 0, 0, 1],
                [1, 0, 1, 0, 1],
                [1, 1, 1, 1, 0],
                [1, 0, 0, 0, 1],
                [100, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [1, 1, 1, 1, 1],
                [0, 0, 0, 0, -50],
            ]
        )
        problem.set_population(pop)
        x = problem.best_child()
        assert np.array_equal(x, np.array([100, 0, 0, 0, 0]))

    def test_best_child_min(self):
        """Test best_child method for a minimization problem"""
        problem = OptProb(5, OneMax(), maximize=False)
        pop = np.array(
            [
                [0, 0, 0, 0, 1],
                [1, 0, 1, 0, 1],
                [1, 1, 1, 1, 0],
                [1, 0, 0, 0, 1],
                [100, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [1, 1, 1, 1, 1],
                [0, 0, 0, 0, -50],
            ]
        )
        problem.set_population(pop)
        x = problem.best_child()
        assert np.array_equal(x, np.array([0, 0, 0, 0, -50]))

    def test_best_neighbor_max(self):
        """Test best_neighbor method for a maximization problem"""
        problem = OptProb(5, OneMax())
        pop = np.array(
            [
                [0, 0, 0, 0, 1],
                [1, 0, 1, 0, 1],
                [1, 1, 1, 1, 0],
                [1, 0, 0, 0, 1],
                [100, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [1, 1, 1, 1, 1],
                [0, 0, 0, 0, -50],
            ]
        )
        problem.neighbors = pop
        x = problem.best_neighbor()
        assert np.array_equal(x, np.array([100, 0, 0, 0, 0]))

    def test_best_neighbor_min(self):
        """Test best_neighbor method for a minimization problem"""
        problem = OptProb(5, OneMax(), maximize=False)
        pop = np.array(
            [
                [0, 0, 0, 0, 1],
                [1, 0, 1, 0, 1],
                [1, 1, 1, 1, 0],
                [1, 0, 0, 0, 1],
                [100, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [1, 1, 1, 1, 1],
                [0, 0, 0, 0, -50],
            ]
        )
        problem.neighbors = pop
        x = problem.best_neighbor()
        assert np.array_equal(x, np.array([0, 0, 0, 0, -50]))

    def test_eval_fitness_max(self):
        """Test eval_fitness method for a maximization problem"""
        problem = OptProb(5, OneMax())
        x = np.array([0, 1, 2, 3, 4])
        fitness = problem.eval_fitness(x)
        assert fitness == 10

    def test_eval_fitness_min(self):
        """Test eval_fitness method for a minimization problem"""
        problem = OptProb(5, OneMax(), maximize=False)
        x = np.array([0, 1, 2, 3, 4])
        fitness = problem.eval_fitness(x)
        assert fitness == -10

    def test_eval_mate_probs(self):
        """Test eval_mate_probs method"""
        problem = OptProb(5, OneMax())
        pop = np.array([[0, 0, 0, 0, 1], [1, 0, 1, 0, 1], [1, 1, 1, 1, 0], [1, 0, 0, 0, 1], [0, 0, 0, 0, 0], [1, 1, 1, 1, 1]])
        problem.set_population(pop)
        problem.eval_mate_probs()
        probs = np.array([0.06667, 0.2, 0.26667, 0.13333, 0, 0.33333])
        assert np.allclose(problem.get_mate_probs(), probs, atol=0.00001)

    def test_eval_mate_probs_maximize_false(self):
        """Test eval_mate_probs method"""
        problem = OptProb(5, OneMax(), maximize=False)
        pop = np.array([[0, 0, 0, 0, 1], [1, 0, 1, 0, 1], [1, 1, 1, 1, 0], [1, 0, 0, 0, 1], [0, 0, 0, 0, 0], [1, 1, 1, 1, 1]])
        problem.set_population(pop)
        problem.eval_mate_probs()
        probs = np.array([0.26667, 0.13333, 0.06667, 0.2, 0.33333, 0])
        assert np.allclose(problem.get_mate_probs(), probs, atol=0.00001)

    def test_eval_mate_probs_all_zero(self):
        """Test eval_mate_probs method when all states have zero fitness"""
        problem = OptProb(5, OneMax())
        pop = np.array([[0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0], [0, 0, 0, 0, 0]])
        problem.set_population(pop)
        problem.eval_mate_probs()
        probs = np.array([0.16667, 0.16667, 0.16667, 0.16667, 0.16667, 0.16667])
        assert np.allclose(problem.get_mate_probs(), probs, atol=0.00001)


class TestDiscreteOpt:
    """Tests for DiscreteOpt class."""

    def test_eval_node_probs(self):
        """Test eval_node_probs method"""
        problem = DiscreteOpt(5, OneMax())
        pop = np.array([[0, 0, 0, 0, 1], [1, 0, 1, 0, 1], [1, 1, 1, 1, 0], [1, 0, 0, 0, 1], [0, 0, 0, 0, 0], [1, 1, 1, 1, 1]])
        problem.keep_sample = pop
        problem.eval_node_probs()
        parent = np.array([2, 0, 1, 0])
        probs = np.array(
            [
                [[0.33333, 0.66667], [0.33333, 0.66667]],
                [[1.0, 0.0], [0.33333, 0.66667]],
                [[1.0, 0.0], [0.25, 0.75]],
                [[1.0, 0.0], [0.0, 1.0]],
                [[0.5, 0.5], [0.25, 0.75]],
            ]
        )
        assert np.allclose(problem.node_probs, probs, atol=0.00001) and np.array_equal(problem.parent_nodes, parent)

    def test_find_neighbors_max2(self):
        """Test find_neighbors method when max_val is equal to 2"""
        problem = DiscreteOpt(5, OneMax())
        x = np.array([0, 1, 0, 1, 0])
        problem.set_state(x)
        problem.find_neighbors()
        neigh = np.array([[1, 1, 0, 1, 0], [0, 0, 0, 1, 0], [0, 1, 1, 1, 0], [0, 1, 0, 0, 0], [0, 1, 0, 1, 1]])
        assert np.array_equal(np.array(problem.neighbors), neigh)

    def test_find_neighbors_max_gt2(self):
        """Test find_neighbors method when max_val is greater than 2"""
        problem = DiscreteOpt(5, OneMax(), max_val=3)
        x = np.array([0, 1, 2, 1, 0])
        problem.set_state(x)
        problem.find_neighbors()
        neigh = np.array(
            [
                [1, 1, 2, 1, 0],
                [2, 1, 2, 1, 0],
                [0, 0, 2, 1, 0],
                [0, 2, 2, 1, 0],
                [0, 1, 0, 1, 0],
                [0, 1, 1, 1, 0],
                [0, 1, 2, 0, 0],
                [0, 1, 2, 2, 0],
                [0, 1, 2, 1, 1],
                [0, 1, 2, 1, 2],
            ]
        )
        assert np.array_equal(np.array(problem.neighbors), neigh)

    def test_find_sample_order(self):
        """Test find_sample_order method"""
        problem = DiscreteOpt(5, OneMax())
        problem.parent_nodes = np.array([2, 0, 1, 0])
        order = np.array([0, 2, 4, 1, 3])
        problem.find_sample_order()
        assert np.array_equal(np.array(problem.sample_order), order)

    def test_find_top_pct_max(self):
        """Test find_top_pct method for a maximization problem"""
        problem = DiscreteOpt(5, OneMax())
        pop = np.array(
            [
                [0, 0, 0, 0, 1],
                [1, 0, 1, 0, 1],
                [1, 1, 1, 1, 0],
                [1, 0, 0, 0, 1],
                [100, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [1, 1, 1, 1, 1],
                [0, 0, 0, 0, -50],
            ]
        )
        problem.set_population(pop)
        problem.find_top_pct(keep_pct=0.25)
        x = np.array([[100, 0, 0, 0, 0], [1, 1, 1, 1, 1]])
        assert np.array_equal(problem.get_keep_sample(), x)

    def test_find_top_pct_min(self):
        """Test find_top_pct method for a minimization problem"""
        problem = DiscreteOpt(5, OneMax(), maximize=False)
        pop = np.array(
            [
                [0, 0, 0, 0, 1],
                [1, 0, 1, 0, 1],
                [1, 1, 1, 1, 0],
                [1, 0, 0, 0, 1],
                [100, 0, 0, 0, 0],
                [0, 0, 0, 0, 0],
                [1, 1, 1, 1, 1],
                [0, 0, 0, 0, -50],
            ]
        )
        problem.set_population(pop)
        problem.find_top_pct(keep_pct=0.25)
        x = np.array([[0, 0, 0, 0, 0], [0, 0, 0, 0, -50]])
        assert np.array_equal(problem.get_keep_sample(), x)

    def test_random(self):
        """Test random method"""
        problem = DiscreteOpt(5, OneMax(), max_val=5)
        rand = problem.random()
        assert len(rand) == 5 and max(rand) >= 0 and min(rand) <= 4

    def test_random_neighbor_max2(self):
        """Test random_neighbor method when max_val is equal to 2"""
        problem = DiscreteOpt(5, OneMax())
        x = np.array([0, 0, 1, 1, 1])
        problem.set_state(x)
        neigh = problem.random_neighbor()
        sum_diff = np.sum(np.abs(x - neigh))
        assert len(neigh) == 5 and sum_diff == 1

    def test_random_neighbor_max_gt2(self):
        """Test random_neighbor method when max_val is greater than 2"""
        problem = DiscreteOpt(5, OneMax(), max_val=5)
        x = np.array([0, 1, 2, 3, 4])
        problem.set_state(x)
        neigh = problem.random_neighbor()
        abs_diff = np.abs(x - neigh)
        abs_diff[abs_diff > 0] = 1
        sum_diff = np.sum(abs_diff)
        assert len(neigh) == 5 and sum_diff == 1

    def test_random_pop(self):
        """Test random_pop method"""
        problem = DiscreteOpt(5, OneMax())
        problem.random_pop(100)
        pop = problem.get_population()
        pop_fitness = problem.get_pop_fitness()
        assert np.shape(pop)[0] == 100 and np.shape(pop)[1] == 5 and 0 < np.sum(pop) < 500 and len(pop_fitness) == 100

    def test_reproduce_mut0(self):
        """Test reproduce method when mutation_prob is 0"""
        problem = DiscreteOpt(5, OneMax())
        father = np.zeros(5)
        mother = np.ones(5)
        child = problem.reproduce(father, mother, mutation_prob=0)
        assert len(child) == 5 and 0 <= sum(child) <= 5

    def test_reproduce_mut1_max2(self):
        """Test reproduce method when mutation_prob is 1 and max_val is 2"""
        problem = DiscreteOpt(5, OneMax())
        father = np.zeros(5)
        mother = np.ones(5)
        child = problem.reproduce(father, mother, mutation_prob=1)
        assert len(child) == 5 and 0 <= sum(child) <= 5

    def test_reproduce_mut1_max_gt2(self):
        """Test reproduce method when mutation_prob is 1 and max_val is greater than 2"""
        problem = DiscreteOpt(5, OneMax(), max_val=3)
        problem._crossover = OnePointCrossover(problem)
        father = np.zeros(5)
        mother = np.ones(5) * 2
        child = problem.reproduce(father, mother, mutation_prob=1)
        assert len(child) == 5 and 0 < sum(child) < 10

    def test_sample_pop(self):
        """Test sample_pop method"""
        problem = DiscreteOpt(5, OneMax())
        pop = np.array([[0, 0, 0, 0, 1], [1, 0, 1, 0, 1], [1, 1, 1, 1, 0], [1, 0, 0, 0, 1], [0, 0, 0, 0, 0], [1, 1, 1, 1, 1]])
        problem.keep_sample = pop
        problem.eval_node_probs()
        sample = problem.sample_pop(100)
        assert np.shape(sample)[0] == 100 and np.shape(sample)[1] == 5 and 0 < np.sum(sample) < 500


class TestContinuousOpt:
    """Tests for ContinuousOpt class."""

    def test_calculate_updates(self):
        """Test calculate_updates method"""
        X = np.array([[0, 1, 0, 1], [0, 0, 0, 0], [1, 1, 1, 1], [1, 1, 1, 1], [0, 0, 1, 1], [1, 0, 0, 0]])
        y = np.reshape(np.array([1, 1, 0, 0, 1, 1]), [6, 1])
        nodes = [4, 2, 1]
        fitness = NetworkWeights(X, y, nodes, activation=identity, bias=False, is_classifier=False, learning_rate=1)

        a = list(np.arange(8) + 1)
        b = list(0.01 * (np.arange(2) + 1))
        weights = a + b
        fitness.evaluate(np.asarray(weights))

        problem = ContinuousOpt(10, fitness, maximize=False)

        updates = problem.calculate_updates()
        update1 = np.array([[-0.0017, -0.0034], [-0.0046, -0.0092], [-0.0052, -0.0104], [0.0014, 0.0028]])
        update2 = np.array([[-3.17], [-4.18]])

        assert np.allclose(updates[0], update1, atol=0.001) and np.allclose(updates[1], update2, atol=0.001)

    def test_find_neighbors_range_eq_step(self):
        """Test find_neighbors method when range equals step size"""
        problem = ContinuousOpt(5, OneMax(), step=1)
        x = np.array([0, 1, 0, 1, 0])
        problem.set_state(x)
        problem.find_neighbors()
        neigh = np.array([[1, 1, 0, 1, 0], [0, 0, 0, 1, 0], [0, 1, 1, 1, 0], [0, 1, 0, 0, 0], [0, 1, 0, 1, 1]])
        assert np.array_equal(np.array(problem.neighbors), neigh)

    def test_find_neighbors_range_gt_step(self):
        """Test find_neighbors method when range greater than step size"""
        problem = ContinuousOpt(5, OneMax(), max_val=2, step=1)
        x = np.array([0, 1, 2, 1, 0])
        problem.set_state(x)
        problem.find_neighbors()
        neigh = np.array(
            [[1, 1, 2, 1, 0], [0, 0, 2, 1, 0], [0, 2, 2, 1, 0], [0, 1, 1, 1, 0], [0, 1, 2, 0, 0], [0, 1, 2, 2, 0], [0, 1, 2, 1, 1]]
        )
        assert np.array_equal(np.array(problem.neighbors), neigh)

    def test_random(self):
        """Test random method"""
        problem = ContinuousOpt(5, OneMax(), max_val=4)
        rand = problem.random()
        assert len(rand) == 5 and max(rand) >= 0 and min(rand) <= 4

    def test_random_neighbor_range_eq_step(self):
        """Test random_neighbor method when range equals step size"""
        problem = ContinuousOpt(5, OneMax(), step=1)
        x = np.array([0, 0, 1, 1, 1])
        problem.set_state(x)
        neigh = problem.random_neighbor()
        sum_diff = np.sum(np.abs(x - neigh))
        assert len(neigh) == 5 and sum_diff == 1

    def test_random_neighbor_range_gt_step(self):
        """Test random_neighbor method when range greater than step size"""
        problem = ContinuousOpt(5, OneMax(), max_val=2, step=1)
        x = np.array([0, 1, 2, 3, 4])
        problem.set_state(x)
        neigh = problem.random_neighbor()
        abs_diff = np.abs(x - neigh)
        abs_diff[abs_diff > 0] = 1
        sum_diff = np.sum(abs_diff)
        assert len(neigh) == 5 and sum_diff == 1

    def test_random_pop(self):
        """Test random_pop method"""
        problem = ContinuousOpt(5, OneMax(), step=1)
        problem.random_pop(100)
        pop = problem.get_population()
        pop_fitness = problem.get_pop_fitness()
        assert np.shape(pop)[0] == 100 and np.shape(pop)[1] == 5 and 0 < np.sum(pop) < 500 and len(pop_fitness) == 100

    def test_reproduce_mut0(self):
        """Test reproduce method when mutation_prob is 0"""
        problem = ContinuousOpt(5, OneMax(), step=1)
        father = np.zeros(5)
        mother = np.ones(5)
        child = problem.reproduce(father, mother, mutation_prob=0)
        assert len(child) == 5 and 0 < sum(child) < 5

    def test_reproduce_mut1_range_eq_step(self):
        """Test reproduce method when mutation_prob is 1 and range equals step size"""
        problem = ContinuousOpt(5, OneMax(), step=1)
        father = np.zeros(5)
        mother = np.ones(5)
        child = problem.reproduce(father, mother, mutation_prob=1)
        assert len(child) == 5 and 0 < sum(child) < 5

    def test_reproduce_mut1_range_gt_step(self):
        """Test reproduce method when mutation_prob is 1 and range is greater than step size"""
        problem = ContinuousOpt(5, OneMax(), max_val=2, step=1)
        father = np.zeros(5)
        mother = np.array([2, 2, 2, 2, 2])
        child = problem.reproduce(father, mother, mutation_prob=1)
        assert len(child) == 5 and 0 < sum(child) < 10

    def test_update_state_in_range(self):
        """Test update_state method where all updated values are within the tolerated range"""
        problem = ContinuousOpt(5, OneMax(), max_val=20, step=1)
        x = np.array([0, 1, 2, 3, 4])
        problem.set_state(x)
        y = np.array([2, 4, 6, 8, 10])
        updated = problem.update_state(y)
        assert np.array_equal(updated, (x + y))

    def test_update_state_outside_range(self):
        """Test update_state method where some updated values are outside the tolerated range"""
        problem = ContinuousOpt(5, OneMax(), max_val=5, step=1)
        x = np.array([0, 1, 2, 3, 4])
        problem.set_state(x)
        y = np.array([2, -4, 6, -8, 10])
        updated = problem.update_state(y)
        z = np.array([2, 0, 5, 0, 5])
        assert np.array_equal(updated, z)


class TestTSPOpt:
    """Tests for TSPOpt class."""

    def test_adjust_probs_all_zero(self):
        """Test adjust_probs method when all elements in input vector sum to zero."""
        dists = [(0, 1, 3), (0, 2, 5), (0, 3, 1), (0, 4, 7), (1, 3, 6), (4, 1, 9), (2, 3, 8), (2, 4, 2), (3, 2, 8), (3, 4, 4)]
        problem = TSPOpt(5, distances=dists)
        probs = np.zeros(5)
        assert np.array_equal(problem.adjust_probs(probs), np.zeros(5))

    def test_adjust_probs_non_zero(self):
        """Test adjust_probs method when all elements in input vector sum to some value other than zero."""
        dists = [(0, 1, 3), (0, 2, 5), (0, 3, 1), (0, 4, 7), (1, 3, 6), (4, 1, 9), (2, 3, 8), (2, 4, 2), (3, 2, 8), (3, 4, 4)]
        problem = TSPOpt(5, distances=dists)
        probs = np.array([0.1, 0.2, 0, 0, 0.5])
        x = np.array([0.125, 0.25, 0, 0, 0.625])
        assert np.array_equal(problem.adjust_probs(probs), x)

    def test_find_neighbors(self):
        """Test find_neighbors method"""
        dists = [(0, 1, 3), (0, 2, 5), (0, 3, 1), (0, 4, 7), (1, 3, 6), (4, 1, 9), (2, 3, 8), (2, 4, 2), (3, 2, 8), (3, 4, 4)]
        problem = TSPOpt(5, distances=dists)
        x = np.array([0, 1, 2, 3, 4])
        problem.set_state(x)
        problem.find_neighbors()
        neigh = np.array(
            [
                [1, 0, 2, 3, 4],
                [2, 1, 0, 3, 4],
                [3, 1, 2, 0, 4],
                [4, 1, 2, 3, 0],
                [0, 2, 1, 3, 4],
                [0, 3, 2, 1, 4],
                [0, 4, 2, 3, 1],
                [0, 1, 3, 2, 4],
                [0, 1, 4, 3, 2],
                [0, 1, 2, 4, 3],
            ]
        )
        assert np.array_equal(np.array(problem.neighbors), neigh)

    def test_random(self):
        """Test random method"""
        dists = [(0, 1, 3), (0, 2, 5), (0, 3, 1), (0, 4, 7), (1, 3, 6), (4, 1, 9), (2, 3, 8), (2, 4, 2), (3, 2, 8), (3, 4, 4)]
        problem = TSPOpt(5, distances=dists)
        rand = problem.random()
        assert len(rand) == 5 and len(set(rand)) == 5

    def test_random_mimic(self):
        """Test random_mimic method"""
        dists = [(0, 1, 3), (0, 2, 5), (0, 3, 1), (0, 4, 7), (1, 3, 6), (4, 1, 9), (2, 3, 8), (2, 4, 2), (3, 2, 8), (3, 4, 4)]
        pop = np.array([[1, 0, 3, 2, 4], [0, 2, 1, 3, 4], [0, 2, 4, 3, 1], [4, 1, 3, 2, 0], [3, 4, 0, 2, 1], [2, 4, 0, 3, 1]])
        problem = TSPOpt(5, distances=dists)
        problem.keep_sample = pop
        problem.eval_node_probs()
        problem.find_sample_order()
        rand = problem.random_mimic()
        assert len(rand) == 5 and len(set(rand)) == 5

    def test_random_neighbor(self):
        """Test random_neighbor method"""
        dists = [(0, 1, 3), (0, 2, 5), (0, 3, 1), (0, 4, 7), (1, 3, 6), (4, 1, 9), (2, 3, 8), (2, 4, 2), (3, 2, 8), (3, 4, 4)]
        problem = TSPOpt(5, distances=dists)
        x = np.array([0, 1, 2, 3, 4])
        problem.set_state(x)
        neigh = problem.random_neighbor()
        abs_diff = np.abs(x - neigh)
        abs_diff[abs_diff > 0] = 1
        sum_diff = np.sum(abs_diff)
        assert len(neigh) == 5 and sum_diff == 2 and len(set(neigh)) == 5

    def test_reproduce_mut0(self):
        """Test reproduce method when mutation_prob is 0"""
        dists = [(0, 1, 3), (0, 2, 5), (0, 3, 1), (0, 4, 7), (1, 3, 6), (4, 1, 9), (2, 3, 8), (2, 4, 2), (3, 2, 8), (3, 4, 4)]
        problem = TSPOpt(5, distances=dists)
        father = np.array([0, 1, 2, 3, 4])
        mother = np.array([0, 4, 3, 2, 1])
        child = problem.reproduce(father, mother, mutation_prob=0)
        assert len(child) == 5 and len(set(child)) == 5

    def test_reproduce_mut1(self):
        """Test reproduce method when mutation_prob is 1"""
        dists = [(0, 1, 3), (0, 2, 5), (0, 3, 1), (0, 4, 7), (1, 3, 6), (4, 1, 9), (2, 3, 8), (2, 4, 2), (3, 2, 8), (3, 4, 4)]
        problem = TSPOpt(5, distances=dists)
        father = np.array([0, 1, 2, 3, 4])
        mother = np.array([4, 3, 2, 1, 0])
        child = problem.reproduce(father, mother, mutation_prob=1)
        assert len(child) == 5 and len(set(child)) == 5

    def test_sample_pop(self):
        """Test sample_pop method"""
        dists = [(0, 1, 3), (0, 2, 5), (0, 3, 1), (0, 4, 7), (1, 3, 6), (4, 1, 9), (2, 3, 8), (2, 4, 2), (3, 2, 8), (3, 4, 4)]
        pop = np.array([[1, 0, 3, 2, 4], [0, 2, 1, 3, 4], [0, 2, 4, 3, 1], [4, 1, 3, 2, 0], [3, 4, 0, 2, 1], [2, 4, 0, 3, 1]])
        problem = TSPOpt(5, distances=dists)
        problem.keep_sample = pop
        problem.eval_node_probs()
        sample = problem.sample_pop(100)
        row_sums = np.sum(sample, axis=1)
        assert np.shape(sample)[0] == 100 and np.shape(sample)[1] == 5 and max(row_sums) == 10 and min(row_sums) == 10



"""Classes for defining neural network weight optimization problems."""

# Author: Genevieve Hayes
# License: BSD 3 clause

import numpy as np
import sklearn.metrics as skm

from mlrose_ky.neural.utils import unflatten_weights
from mlrose_ky.neural import activation as act


class NetworkWeights:
    """Fitness function for neural network weights optimization problem.

    Parameters
    ----------
    X: np.ndarray
        Numpy array containing feature dataset with each row representing a
        single observation.

    y: np.ndarray
        Numpy array containing true values of data labels.
        Length must be same as length of X.

    node_list: list[int]
        Number of nodes in each layer, including the input and output layers.

    activation: callable
        Activation function for each of the hidden layers with the signature
        :code:`activation(x, deriv)`, where setting deriv is a boolean that
        determines whether to return the activation function or its derivative.

    bias: bool, default: True
        Whether a bias term is included in the network.

    is_classifier: bool, default: True
        Whether the network is for classification or regression. Set True for
        classification and False for regression.
    """

    def __init__(self, X, y, node_list: list[int], activation, bias=True, is_classifier=True, learning_rate=0.1):

        # Make sure y is an array and not a list
        y = np.array(y)

        # Convert y to 2D if necessary
        if len(np.shape(y)) == 1:
            y = np.reshape(y, [len(y), 1])

        # Verify X and y are the same length
        if not np.shape(X)[0] == np.shape(y)[0]:
            raise Exception("""The length of X and y must be equal.""")

        if len(node_list) < 2:
            raise Exception("""node_list must contain at least 2 elements.""")

        if not np.shape(X)[1] == (node_list[0] - bias):
            raise Exception("""The number of columns in X must equal %d""" % ((node_list[0] - bias),))

        if not np.shape(y)[1] == node_list[-1]:
            raise Exception("""The number of columns in y must equal %d""" % (node_list[-1],))

        if not isinstance(bias, bool):
            raise Exception("""bias must be True or False.""")

        if not isinstance(is_classifier, bool):
            raise Exception("""is_classifier must be True or False.""")

        if learning_rate <= 0:
            raise Exception("""learning_rate must be greater than 0.""")

        self.X = X
        self.y_true = y
        self.node_list = node_list
        self.activation = activation
        self.bias = bias
        self.is_classifier = is_classifier
        self.learning_rate = learning_rate

        # Determine appropriate loss function and output activation function
        if self.is_classifier:
            self.loss = skm.log_loss

            if np.shape(self.y_true)[1] == 1:
                self.output_activation = act.sigmoid
            else:
                self.output_activation = act.softmax
        else:
            self.loss = skm.mean_squared_error
            self.output_activation = act.identity

        self.inputs_list = []
        self.y_pred = y
        self.weights = []
        self.prob_type = "continuous"

        nodes = 0
        for i in range(len(node_list) - 1):
            nodes += node_list[i] * node_list[i + 1]

        self.nodes = nodes

    def evaluate(self, state):
        """Evaluate the fitness of a state.

        Parameters
        ----------
        state: np.ndarray
            State array for evaluation.

        Returns
        -------
        fitness: float
            Value of fitness function.
        """
        if not len(state) == self.nodes:
            raise Exception("""state must have length %d""" % (self.nodes,))

        self.inputs_list = []
        self.weights: list = list(unflatten_weights(state, self.node_list))

        # Add bias column to inputs matrix, if required
        if self.bias:
            ones = np.ones([np.shape(self.X)[0], 1])
            inputs = np.hstack((self.X, ones))

        else:
            inputs = self.X

        # Pass data through network
        for i in range(len(self.weights)):
            # Multiple inputs by weights
            outputs = np.dot(inputs, self.weights[i])
            self.inputs_list.append(inputs)

            # Transform outputs to get inputs for next layer (or final preds)
            if i < len(self.weights) - 1:
                inputs = self.activation(outputs)
            else:
                self.y_pred = self.output_activation(outputs)

        # Evaluate loss function
        fitness = self.loss(self.y_true, self.y_pred)

        return fitness

    def get_output_activation(self):
        """Return the activation function for the output layer.

        Returns
        -------
        self.output_activation: callable
            Activation function for the output layer.
        """
        return self.output_activation

    def get_problem_type(self):
        """Return the problem type.

        Returns
        -------
        self.prob_type: str
            Specifies problem type as 'discrete', 'continuous', 'tsp', or
            'either'.
        """
        return self.prob_type

    def calculate_updates(self):
        """Calculate gradient descent updates.

        Returns
        -------
        updates_list: list
            List of back propagation weight updates.
        """
        delta_list: list = []
        updates_list = []

        # Work backwards from final layer
        for i in range(len(self.inputs_list) - 1, -1, -1):
            # Final layer
            if i == len(self.inputs_list) - 1:
                delta = self.y_pred - self.y_true
            # Hidden layers
            else:
                dot = np.dot(delta_list[-1], np.transpose(self.weights[i + 1]))
                activation = self.activation(self.inputs_list[i + 1], deriv=True)
                delta = dot * activation

            delta_list.append(delta)

            # Calculate updates
            updates = -1.0 * self.learning_rate * np.dot(np.transpose(self.inputs_list[i]), delta)

            updates_list.append(updates)

        # Reverse order of updates list
        updates_list = updates_list[::-1]

        return updates_list



"""Classes for defining neural network weight optimization problems."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

from sklearn.base import RegressorMixin


from mlrose_ky.algorithms.decay import GeometricDecay
from .nn_core import _NNCore


class LinearRegression(_NNCore, RegressorMixin):
    """Class for defining linear regression weights optimization
    problem. Inherits :code:`fit` and :code:`predict` methods from
    :code:`NeuralNetwork()` class.

    Parameters
    ----------
    algorithm: str, default: 'random_hill_climb'
        Algorithm used to find optimal network weights. Must be one
        of:'random_hill_climb', 'simulated_annealing', 'genetic_alg' or
        'gradient_descent'.

    max_iters: int, default: 100
        Maximum number of iterations used to fit the weights.

    bias: bool, default: True
        Whether to include a bias term.

    learning_rate: float, default: 0.1
        Learning rate for gradient descent or step size for randomized
        optimization algorithms.

    early_stopping: bool, default: False
        Whether to terminate algorithm early if the loss is not improving.
        If :code:`True`, then stop after max_attempts iters with no
        improvement.

    clip_max: float, default: 1e+10
        Used to limit weights to the range [-1*clip_max, clip_max].

    restarts: int, default: 0
        Number of random restarts.
        Only required if :code:`algorithm = 'random_hill_climb'`.

    schedule: schedule object, default = mlrose_ky.GeometricDecay()
        Schedule used to determine the value of the temperature parameter.
        Only required if :code:`algorithm = 'simulated_annealing'`.

    pop_size: int, default: 200
        Size of population. Only required if :code:`algorithm = 'genetic_alg'`.

    mutation_prob: float, default: 0.1
        Probability of a mutation at each element of the state vector during
        reproduction, expressed as a value between 0 and 1. Only required if
        :code:`algorithm = 'genetic_alg'`.

    max_attempts: int, default: 10
        Maximum number of attempts to find a better state. Only required if
        :code:`early_stopping = True`.

    random_state: int, default: None
        If random_state is a positive integer, random_state is the seed used
        by np.random.seed(); otherwise, the random seed is not set.

    curve: bool, default: False
        If bool is true, curve containing the fitness at each training
        iteration is returned.

    Attributes
    ----------
    fitted_weights: np.ndarray
        Numpy array giving the fitted weights when :code:`fit` is performed.

    loss: float
        Value of loss function for fitted weights when :code:`fit` is
        performed.

    fitness_curve: np.ndarray
        Numpy array giving the fitness at each training iteration.
    """

    def __init__(
        self,
        algorithm="random_hill_climb",
        max_iters=100,
        bias=True,
        learning_rate=0.1,
        early_stopping=False,
        clip_max=1e10,
        restarts=0,
        schedule=GeometricDecay(),
        pop_size=200,
        mutation_prob=0.1,
        max_attempts=10,
        random_state=None,
        curve=False,
    ):
        _NNCore.__init__(
            self,
            hidden_nodes=[],
            activation="identity",
            algorithm=algorithm,
            max_iters=max_iters,
            bias=bias,
            is_classifier=False,
            learning_rate=learning_rate,
            early_stopping=early_stopping,
            clip_max=clip_max,
            restarts=restarts,
            schedule=schedule,
            pop_size=pop_size,
            mutation_prob=mutation_prob,
            max_attempts=max_attempts,
            random_state=random_state,
            curve=curve,
        )



"""Classes for defining optimization problem objects."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

import numpy as np

from mlrose_ky.algorithms.crossovers import UniformCrossover
from mlrose_ky.algorithms.mutators import SingleGeneMutator
from mlrose_ky.fitness.queens import Queens
from mlrose_ky.opt_probs.discrete_opt import DiscreteOpt


class QueensOpt(DiscreteOpt):
    def __init__(self, length=None, fitness_fn=None, maximize=False, crossover=None, mutator=None):

        if (fitness_fn is None) and (length is None):
            raise Exception("fitness_fn or length must be specified.")

        if length is None:
            length = len(fitness_fn.weights)

        self.length = length

        if fitness_fn is None:
            fitness_fn = Queens(maximize=maximize)

        self.stop_fitness = Queens.get_max_size(length) if maximize else 0

        self.max_val = length
        crossover = UniformCrossover(self) if crossover is None else crossover
        mutator = SingleGeneMutator(self) if mutator is None else mutator
        super().__init__(length, fitness_fn, maximize, length, crossover, mutator)

        state = np.random.randint(self.length, size=self.length)
        np.random.shuffle(state)
        self.set_state(state)

    def can_stop(self):
        return int(self.get_fitness()) == self.stop_fitness



"""
GridSearchMixin: A mixin providing grid search functionality for a neural network runner.

This mixin allows for parameter optimization through grid search, leveraging scikit-learn's
GridSearchCV for cross-validated search over parameter grids.
"""

import inspect
import numpy as np
import sklearn.metrics as skmt
import sklearn.model_selection as skms
from typing import Callable, Any


class GridSearchMixin:
    """
    A mixin class providing grid search capabilities using scikit-learn's GridSearchCV.

    Attributes
    ----------
    scorer_method : Callable
        The scoring method used for evaluating model performance during grid search.
    params : inspect.Signature
        The signature of the scoring method, used for validation and handling of additional arguments.
    get_y_argmax : bool
        Flag indicating whether to apply argmax to predictions and ground truths before scoring.
    """

    def __init__(self, scorer_method: Callable = None) -> None:
        """
        Initializes the GridSearchMixin with a specified scoring method.

        Parameters
        ----------
        scorer_method : Callable, optional
            A custom scoring method for evaluating the model. Defaults to balanced accuracy.
        """
        self.scorer_method: Callable = skmt.balanced_accuracy_score if scorer_method is None else scorer_method
        self.params = inspect.signature(self.scorer_method)
        self.get_y_argmax: bool = False

    def perform_grid_search(
        self,
        classifier: Any,
        x_train: np.ndarray,
        y_train: np.ndarray,
        cv: int,
        parameters: dict,
        n_jobs: int = 1,
        verbose: bool = False
    ) -> skms.GridSearchCV:
        """
        Perform grid search with cross-validation on the provided classifier.

        Parameters
        ----------
        classifier : Any
            The machine learning model or pipeline to optimize.
        x_train : np.ndarray
            Training data features.
        y_train : np.ndarray
            Training data labels.
        cv : int
            Number of cross-validation folds.
        parameters : dict
            Dictionary with parameters names as keys and lists of parameter settings to try as values.
        n_jobs : int, optional
            Number of jobs to run in parallel. Defaults to 1.
        verbose : bool, optional
            Whether to display verbose output during grid search. Defaults to False.

        Returns
        -------
        skms.GridSearchCV
            The fitted GridSearchCV object containing the results of the grid search.
        """
        scorer = self.make_scorer()
        search_results = skms.GridSearchCV(
            classifier, parameters, cv=cv, scoring=scorer, n_jobs=n_jobs, return_train_score=True, verbose=verbose
        )
        search_results.fit(x_train, y_train)
        return search_results

    def make_scorer(self) -> Callable:
        """
        Create a custom scorer function using the provided scoring method.

        Returns
        -------
        Callable
            A scorer callable suitable for use with scikit-learn's GridSearchCV.
        """
        return skmt.make_scorer(self._grid_search_score_intercept)

    def score(self, **kwargs: Any) -> float:
        """
        Calculate the score using the scoring method.

        Parameters
        ----------
        kwargs : dict
            Additional arguments to pass to the scoring method.

        Returns
        -------
        float
            The score calculated by the scoring method.
        """
        return self._grid_search_score_intercept(**kwargs)

    def _grid_search_score_intercept(self, y_pred: np.ndarray, y_true: np.ndarray, **kwargs: Any) -> float:
        """
        Intercept method for handling grid search scoring, accommodating special cases for prediction and true values.

        Parameters
        ----------
        y_pred : np.ndarray
            The predicted labels.
        y_true : np.ndarray
            The true labels.
        kwargs : dict
            Additional arguments to pass to the scoring method.

        Returns
        -------
        float
            The calculated score based on the predictions and true values.
        """
        cleaned_kwargs = {k: v for k, v in kwargs.items() if k in self.params.parameters}

        # Handle potential multi-class/multi-label cases
        if not self.get_y_argmax and y_pred.ndim > 1 and y_true.ndim > 1:
            try:
                return self.scorer_method(y_pred=y_pred, y_true=y_true, **cleaned_kwargs)
            except TypeError:
                self.get_y_argmax = True

        if self.get_y_argmax:
            y_pred = y_pred.argmax(axis=1)
            y_true = y_true.argmax(axis=1)

        try:
            return self.scorer_method(y_pred=y_pred, y_true=y_true, **cleaned_kwargs)
        except TypeError as e:
            raise ValueError(f"Error during scoring with method {self.scorer_method.__name__}: {e}") from e



"""Classes for running optimization problems."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

from .ga_runner import GARunner
from .rhc_runner import RHCRunner
from .sa_runner import SARunner
from .mimic_runner import MIMICRunner
from .nngs_runner import NNGSRunner
from .skmlp_runner import SKMLPRunner
from .utils import build_data_filename



"""Class defining a Knapsack optimization problem generator."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

import numpy as np
from mlrose_ky import KnapsackOpt


class KnapsackGenerator:
    """A class to generate Knapsack optimization problems."""

    @staticmethod
    def generate(
        seed: int,
        number_of_item_types: int = 10,
        max_item_count: int = 5,
        max_weight_per_item: int = 25,
        max_value_per_item: int = 10,
        max_weight_percentage: float = 0.6,
        multiply_by_max_item_count: bool = True,
    ) -> KnapsackOpt:
        """
        Generate a Knapsack optimization problem instance.

        Parameters
        ----------
        seed : int
            Seed for the random number generator.
        number_of_item_types : int, optional, default=10
            Number of different item types.
        max_item_count : int, optional, default=5
            Maximum count for each item type.
        max_weight_per_item : int, optional, default=25
            Maximum weight for each item type.
        max_value_per_item : int, optional, default=10
            Maximum value for each item type.
        max_weight_percentage : float, optional, default=0.6
            Maximum weight percentage of the knapsack.
        multiply_by_max_item_count : bool, optional, default=True
            If True, multiply weights and values by max_item_count.

        Returns
        -------
        KnapsackOpt
            An instance of KnapsackOpt configured with the specified parameters.

        Raises
        ------
        ValueError
            If any parameter is not of the expected type or value.
        """
        if not isinstance(seed, int):
            raise ValueError(f"Seed must be an integer. Got {seed}")
        if not isinstance(number_of_item_types, int) or number_of_item_types <= 0:
            raise ValueError(f"Number of item types must be a positive integer. Got {number_of_item_types}")
        if not isinstance(max_item_count, int) or max_item_count <= 0:
            raise ValueError(f"Max item count must be a positive integer. Got {max_item_count}")
        if not isinstance(max_weight_per_item, int) or max_weight_per_item <= 0:
            raise ValueError(f"Max weight per item must be a positive integer. Got {max_weight_per_item}")
        if not isinstance(max_value_per_item, int) or max_value_per_item <= 0:
            raise ValueError(f"Max value per item must be a positive integer. Got {max_value_per_item}")
        if not isinstance(max_weight_percentage, float) or not (0 <= max_weight_percentage <= 1):
            raise ValueError(f"Max weight percentage must be a float between 0 and 1. Got {max_weight_percentage}")
        if not isinstance(multiply_by_max_item_count, bool):
            raise ValueError(f"multiply_by_max_item_count must be a boolean. Got {multiply_by_max_item_count}")

        np.random.seed(seed)

        weights = 1 + np.random.randint(max_weight_per_item, size=number_of_item_types)
        values = 1 + np.random.randint(max_value_per_item, size=number_of_item_types)

        problem = KnapsackOpt(
            length=number_of_item_types,
            max_val=max_item_count,
            weights=weights,
            values=values,
            max_weight_pct=max_weight_percentage,
            multiply_by_max_item_count=multiply_by_max_item_count,
        )

        return problem



"""Unit tests for neural/fitness/"""

# Author: Genevieve Hayes (modified by Kyle Nakamura)
# License: BSD 3 clause

import numpy as np

try:
    import mlrose_ky
except ImportError:
    import sys

    sys.path.append("..")

from mlrose_ky import (
    OneMax,
    FlipFlop,
    FourPeaks,
    SixPeaks,
    ContinuousPeaks,
    Knapsack,
    TravellingSalesperson,
    Queens,
    MaxKColor,
    CustomFitness,
)
from mlrose_ky.fitness._discrete_peaks_base import _DiscretePeaksBase


def test_onemax():
    """Test OneMax fitness function"""
    state = np.array([0, 1, 0, 1, 1, 1, 1])
    assert OneMax().evaluate(state) == 5


def test_flipflop():
    """Test FlipFlop fitness function"""
    state = np.array([0, 1, 0, 1, 1, 1, 1])
    assert FlipFlop().evaluate(state) == 3


def test_head():
    """Test head function"""
    state = np.array([1, 1, 1, 1, 0, 1, 0, 2, 1, 1, 1, 1, 1, 4, 6, 1, 1])
    assert _DiscretePeaksBase.count_leading_values(1, state) == 4


def test_tail():
    """Test tail function"""
    state = np.array([1, 1, 1, 1, 0, 1, 0, 2, 1, 1, 1, 1, 1, 4, 6, 1, 1])
    assert _DiscretePeaksBase.count_trailing_values(1, state) == 2


def test_max_run_middle():
    """Test max_run function for case where run is in the middle of the state"""
    state = np.array([1, 1, 1, 1, 0, 1, 0, 2, 1, 1, 1, 1, 1, 4, 6, 1, 1])
    assert ContinuousPeaks._max_run(1, state) == 5


def test_max_run_start():
    """Test max_run function for case where run is at the start of the state"""
    state = np.array([1, 1, 1, 1, 1, 1, 0, 2, 1, 1, 1, 1, 1, 4, 6, 1, 1])
    assert ContinuousPeaks._max_run(1, state) == 6


def test_max_run_end():
    """Test max_run function for case where run is at the end of the state"""
    state = np.array([1, 1, 1, 1, 0, 1, 0, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1])
    assert ContinuousPeaks._max_run(1, state) == 9


def test_fourpeaks_r0():
    """Test FourPeaks fitness function for the case where R=0 and max>0"""
    state = np.array([1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0])
    assert FourPeaks(threshold_percentage=0.30).evaluate(state) == 4


def test_fourpeaks_r_gt0():
    """Test FourPeaks fitness function for the case where R>0 and max>0"""
    state = np.array([1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0])
    assert FourPeaks(threshold_percentage=0.15).evaluate(state) == 16


def test_fourpeaks_r0_max0():
    """Test FourPeaks fitness function for the case where R=0 and max=0"""
    state = np.array([0, 0, 0, 1, 0, 1, 1, 0, 1, 1, 1, 1])
    assert FourPeaks(threshold_percentage=0.30).evaluate(state) == 0


def test_sixpeaks_r0():
    """Test SixPeaks fitness function for the case where R=0 and max>0"""
    state = np.array([1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0])
    assert SixPeaks(threshold_percentage=0.30).evaluate(state) == 4


def test_sixpeaks_r_gt0():
    """Test SixPeaks fitness function for the case where R>0 and max>0"""
    state = np.array([1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 0, 0])
    assert SixPeaks(threshold_percentage=0.15).evaluate(state) == 16


def test_sixpeaks_r0_max0():
    """Test SixPeaks fitness function for the case where R=0 and max=0"""
    state = np.array([0, 0, 0, 1, 0, 1, 1, 0, 1, 1, 1, 1])
    assert SixPeaks(threshold_percentage=0.30).evaluate(state) == 0


def test_sixpeaks_r_gt0_max0():
    """Test SixPeaks fitness function for the case where R>0 and max=0"""
    state = np.array([0, 0, 0, 1, 0, 1, 1, 0, 1, 1, 1, 1])
    assert SixPeaks(threshold_percentage=0.15).evaluate(state) == 12


def test_continuouspeaks_r0():
    """Test ContinuousPeaks fitness function for case when R = 0."""
    state = np.array([0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1])
    assert ContinuousPeaks(threshold_percentage=0.30).evaluate(state) == 5


def test_continuouspeaks_r_gt():
    """Test ContinuousPeaks fitness function for case when R > 0."""
    state = np.array([0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1])
    assert ContinuousPeaks(threshold_percentage=0.15).evaluate(state) == 17


def test_knapsack_weight_lt_max():
    """Test Knapsack fitness function for case where total weight is less than the maximum"""
    weights = [10, 5, 2, 8, 15]
    values = [1, 2, 3, 4, 5]
    max_weight_pct = 0.6
    state = np.array([1, 0, 2, 1, 0])
    calculated_weights = Knapsack(weights, values, max_weight_pct).evaluate(state)
    assert calculated_weights == 11


def test_knapsack_weight_gt_max():
    """Test Knapsack fitness function for case where total weight is greater than the maximum"""
    weights = [10, 5, 2, 8, 15]
    values = [1, 2, 3, 4, 5]
    max_weight_pct = 0.4
    state = np.array([1, 0, 2, 1, 0])
    assert Knapsack(weights, values, max_weight_pct).evaluate(state) == 0


def test_travelling_sales_coords():
    """Test TravellingSales fitness function for case where city nodes coords are specified."""
    coords = [(0, 0), (3, 0), (3, 2), (2, 4), (1, 3)]
    state = np.array([0, 1, 4, 3, 2])
    assert round(TravellingSalesperson(coords=coords).evaluate(state), 4) == 13.8614


def test_travelling_sales_dists():
    """Test TravellingSales fitness function for case where distances between node pairs are specified."""
    dists = [(0, 1, 3), (0, 2, 5), (0, 3, 1), (0, 4, 7), (1, 3, 6), (4, 1, 9), (2, 3, 8), (2, 4, 2), (3, 2, 8), (3, 4, 4)]
    state = np.array([0, 1, 4, 3, 2])
    assert TravellingSalesperson(distances=dists).evaluate(state) == 29


def test_travelling_sales_invalid():
    """Test TravellingSales fitness function for invalid tour"""
    dists = [(0, 1, 3), (0, 2, 5), (0, 3, 1), (0, 4, 7), (1, 3, 6), (4, 1, 9), (2, 3, 8), (2, 4, 2), (3, 2, 8), (3, 4, 4)]
    state = np.array([0, 1, 2, 3, 4])
    assert TravellingSalesperson(distances=dists).evaluate(state) == np.inf


def test_queens():
    """Test Queens fitness function"""
    state = np.array([1, 4, 1, 3, 5, 5, 2, 7])
    fitness = Queens().evaluate(state)
    assert fitness == 6


def test_max_k_color():
    """Test MaxKColor fitness function"""
    edges = [(0, 1), (0, 2), (0, 4), (1, 3), (2, 0), (2, 3), (3, 4), (0, 5)]
    state = np.array([0, 1, 0, 1, 1, 1])
    assert MaxKColor(edges).evaluate(state) == 3


def test_custom_fitness():
    """Test CustomFitness fitness function"""

    def custom_fitness(_state, c):
        return c * np.sum(_state)

    state = np.array([1, 2, 3, 4, 5])
    kwargs = {"c": 10}
    assert CustomFitness(custom_fitness, **kwargs).evaluate(state) == 150



"""One Point Crossover implementation for Genetic Algorithms (GA).

This module defines a one-point crossover operation used in genetic algorithms,
where a single crossover point is chosen randomly to combine two parent solutions.
"""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 Clause

import numpy as np
from typing import Any, Sequence

from mlrose_ky.algorithms.crossovers._crossover_base import _CrossoverBase


class OnePointCrossover(_CrossoverBase):
    """
    One-point crossover for genetic algorithms.

    This class implements a simple one-point crossover, where a single crossover
    point on the parent chromosomes is chosen randomly, and the genetic information
    is exchanged to create a new offspring.

    Inherits from:
    _CrossoverBase : Abstract base class for crossover operations.
    """

    def __init__(self, optimization_problem: Any) -> None:
        """
        Initialize the OnePointCrossover with the given optimization problem.

        Parameters
        ----------
        optimization_problem : Any
            An instance of the optimization problem related to the genetic algorithm.
        """
        super().__init__(optimization_problem)

    def mate(self, parent1: Sequence[float], parent2: Sequence[float]) -> np.ndarray:
        """
        Perform the one-point crossover between two parent sequences to produce offspring.

        A single crossover point is selected randomly from the chromosome. All genes before
        that point are copied from the first parent and the rest from the second parent.

        Parameters
        ----------
        parent1 : Sequence[float]
            The first parent chromosome sequence.
        parent2 : Sequence[float]
            The second parent chromosome sequence.

        Returns
        -------
        np.ndarray
            The offspring chromosome resulting from the crossover.
        """
        crossover_point = 1 + np.random.randint(self.chromosome_length - 1)
        offspring = np.array([*parent1[:crossover_point], *parent2[crossover_point:]])
        return offspring



"""Crossover implementations for Genetic Algorithms (GA).

This module defines a base class for crossover operations used in genetic algorithms,
detailing how two parent solutions can be combined to create offspring.
"""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 Clause

from abc import ABC, abstractmethod
from typing import Any


class _CrossoverBase(ABC):
    """
    Base class for crossover operations in a genetic algorithm.

    Provides a structured way to define crossover behavior in genetic algorithms.
    It should be subclassed to implement specific crossover strategies.

    Parameters
    ----------
    optimization_problem : Any
        An instance of the optimization problem related to the genetic algorithm.
        This problem instance should provide necessary properties like 'length'
        that might be needed for the crossover operation.

    Attributes
    ----------
    optimization_problem : Any
        The optimization problem instance.
    chromosome_length : int
        Length of the chromosome, typically derived from the optimization problem's
        'length' property.
    """

    def __init__(self, optimization_problem: Any) -> None:
        """
        Initialize the CrossoverBase with the given optimization problem.

        Parameters
        ----------
        optimization_problem : Any
            An instance of the optimization problem related to the GA.
        """
        super().__init__()
        self.optimization_problem = optimization_problem
        self.chromosome_length: int = optimization_problem.length

    @abstractmethod
    def mate(self, parent1: Any, parent2: Any) -> Any:
        """
        Perform the crossover (mating) between two parents to produce offspring.

        This method must be implemented by subclasses to define specific crossover
        behavior based on the genetics of the parents.

        Parameters
        ----------
        parent1 : Any
            The first parent participating in the crossover.
        parent2 : Any
            The second parent participating in the crossover.

        Returns
        -------
        Any
            The offspring resulting from the crossover. The type of this result
            can vary depending on the specific GA implementation.
        """
        pass



"""Classes for defining neural network weight optimization problems."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

import numpy as np


def flatten_weights(weights: list) -> np.ndarray:
    """Flatten list of weights arrays into a 1D array.

    Parameters
    ----------
    weights: list
        List of 2D arrays for flattening.

    Returns
    -------
    flat_weights: np.ndarray
        1D weights array.
    """
    flat_weights = []

    for i in range(len(weights)):
        flat_weights += list(weights[i].flatten())

    flat_weights = np.array(flat_weights)

    return flat_weights


def unflatten_weights(flat_weights: np.ndarray, node_list: list) -> list:
    """Convert 1D weights array into list of 2D arrays.

    Parameters
    ----------
    flat_weights: np.ndarray
        1D weights array.

    node_list: list
        List giving the number of nodes in each layer of the network,
        including the input and output layers.

    Returns
    -------
    weights: list
        List of 2D arrays created from flat_weights.
    """
    nodes = 0
    for i in range(len(node_list) - 1):
        nodes += node_list[i] * node_list[i + 1]

    if len(flat_weights) != nodes:
        raise Exception("""flat_weights must have length %d""" % (nodes,))

    weights = []
    start = 0

    for i in range(len(node_list) - 1):
        end = start + node_list[i] * node_list[i + 1]
        weights.append(np.reshape(flat_weights[start:end], [node_list[i], node_list[i + 1]]))
        start = end

    return weights


def gradient_descent_original(problem, max_attempts=10, max_iters=np.inf, init_state=None, curve=False, random_state=None):
    """Use gradient_descent to find the optimal neural network weights.
    Parameters
    ----------
    problem: optimization object
        Object containing optimization problem to be solved.
    max_attempts: int, default: 10
        Maximum number of attempts to find a better state at each step.
    max_iters: int, default: np.inf
        Maximum number of iterations of the algorithm.
    init_state: np.ndarray, default: None
        Numpy array containing starting state for algorithm.
        If None, then a random state is used.
    random_state: int, default: None
        If random_state is a positive integer, random_state is the seed used
        by np.random.seed(); otherwise, the random seed is not set.
    curve: bool, default: False
        Boolean to keep fitness values for a curve.
        If :code:`False`, then no curve is stored.
        If :code:`True`, then a history of fitness values is provided as a
        third return value.
    Returns
    -------
    best_state: np.ndarray
        Numpy array containing state that optimizes fitness function.
    best_fitness: float
        Value of fitness function at best state.
    fitness_curve: np.ndarray
        Numpy array containing the fitness at every iteration.
        Only returned if input argument :code:`curve` is :code:`True`.
    """
    if (not isinstance(max_attempts, int) and not max_attempts.is_integer()) or max_attempts < 0:
        raise Exception("""max_attempts must be a positive integer.""")

    if (not isinstance(max_iters, int) and max_iters != np.inf and not max_iters.is_integer()) or max_iters < 0:
        raise Exception("""max_iters must be a positive integer.""")

    if init_state is not None and len(init_state) != problem.get_length():
        raise Exception("""init_state must have same length as problem.""")

    # Set random seed
    if isinstance(random_state, int) and random_state > 0:
        np.random.seed(random_state)

    # Initialize problem, time and attempts counter
    if init_state is None:
        problem.reset()
    else:
        problem.set_state(init_state)

    fitness_curve = []
    attempts = 0
    iters = 0

    best_fitness = problem.get_maximize() * problem.get_fitness()
    best_state = problem.get_state()

    while (attempts < max_attempts) and (iters < max_iters):
        iters += 1

        # Update weights
        updates = flatten_weights(problem.calculate_updates())

        next_state = problem.update_state(updates)
        next_fitness = problem.eval_fitness(next_state)

        if next_fitness > problem.get_fitness():
            attempts = 0
        else:
            attempts += 1

        if next_fitness > problem.get_maximize() * best_fitness:
            best_fitness = problem.get_maximize() * next_fitness
            best_state = next_state

        if curve:
            fitness_curve.append(problem.get_adjusted_fitness())

        problem.set_state(next_state)

    if curve:
        return best_state, best_fitness, np.asarray(fitness_curve)

    return best_state, best_fitness, None



"""Single Gene (Change One) Mutator for Genetic Algorithms (GA)."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

import numpy as np
from typing import Any

from mlrose_ky.algorithms.mutators._mutator_base import _MutatorBase


class SingleGeneMutator(_MutatorBase):
    """
    A mutator class that performs the 'Change One' mutation strategy in a genetic algorithm.

    This mutator changes one randomly selected gene in the chromosome to a new value within the allowed range.

    Attributes
    ----------
    optimization_problem : Any
        The optimization problem instance associated with the mutation operations.
    max_gene_value : int
        The maximum allowable value for any gene in the chromosome.

    Parameters
    ----------
    optimization_problem : Any
        An instance of an optimization problem that the mutator will operate on.
    """

    def __init__(self, optimization_problem: Any) -> None:
        super().__init__(optimization_problem)
        self.max_gene_value = optimization_problem.max_val

    def mutate(self, child: np.ndarray, mutation_probability: float) -> np.ndarray:
        """
        Apply 'Change One' mutation operation to a given child chromosome based on a mutation probability.

        Randomly selects one gene and changes its value to a random value within the permissible range.

        Parameters
        ----------
        child : np.ndarray
            The chromosome of a child individual to be mutated.
        mutation_probability : float
            The probability of the child undergoing mutation.

        Returns
        -------
        np.ndarray
            The mutated chromosome.

        Raises
        ------
        ValueError
            If the mutation_probability is not within the range [0, 1].
        """
        if not (0 <= mutation_probability <= 1):
            raise ValueError(f"Mutation probability must be between 0 and 1, got {mutation_probability}")

        if np.random.rand() < mutation_probability:
            mutation_index = np.random.randint(len(child))
            child[mutation_index] = np.random.randint(self.max_gene_value)

        return child



import os


def build_data_filename(output_directory, runner_name, experiment_name, df_name, x_param="", y_param="", ext=""):
    # ensure directory exists
    try:
        os.makedirs(os.path.join(output_directory, experiment_name))
    except (OSError, IOError, FileExistsError, Exception):
        pass

    # return full filename
    if len(ext) > 0 and not ext[0] == ".":
        ext = f".{ext}"
    if len(x_param) > 0 and not x_param[0] == "_" and not x_param[-1] == "_":
        x_param = f"_{x_param}_"
    if len(y_param) > 0 and not y_param[0] == "_" and not y_param[-1] == "_":
        y_param = f"_{y_param}"
    return os.path.join(output_directory, experiment_name, f"{runner_name.lower()}__{experiment_name}__{df_name}{x_param}{y_param}{ext}")



"""Class defining the Flip Flop fitness function for use with optimization algorithms."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

import numpy as np


class FlipFlop:
    """Fitness function for Flip Flop optimization problem.

    Evaluates the fitness of a state vector :math:`x` as the total number of pairs of consecutive
    elements of :math:`x`, (:math:`x_{i}` and :math:`x_{i+1}`) where :math:`x_{i} \\neq x_{i+1}`.

    Examples
    --------
    >>> fitness = FlipFlop()
    >>> state_vector = np.array([0, 1, 0, 1, 1, 1, 1])
    >>> fitness.evaluate(state_vector)
    3.0

    Note
    ----
    The Flip Flop fitness function is suitable for use in discrete-state
    optimization problems *only*.
    """

    def __init__(self):
        """Initialize the Flip Flop fitness function."""
        self.problem_type: str = "discrete"

    @staticmethod
    def evaluate(state_vector: np.ndarray) -> float:
        """Evaluate the fitness of a state vector.

        Parameters
        ----------
        state_vector : np.ndarray
            State array for evaluation.

        Returns
        -------
        float
            Value of fitness function.

        Raises
        ------
        TypeError
            If `state_vector` is not an instance of `np.ndarray`.
        """
        if not isinstance(state_vector, np.ndarray):
            raise TypeError(f"Expected state_vector to be np.ndarray, got {type(state_vector).__name__} instead.")

        differences = np.diff(state_vector) != 0
        fitness = float(np.sum(differences))
        return fitness

    @staticmethod
    def evaluate_many(state_matrix: np.ndarray) -> np.ndarray:
        """Evaluate the fitness of an ndarray of state vectors.

        Parameters
        ----------
        state_matrix : np.ndarray
            States array for evaluation.

        Returns
        -------
        np.ndarray
            Array of fitness values.

        Raises
        ------
        TypeError
            If `state_matrix` is not an instance of `np.ndarray`.
        """
        if not isinstance(state_matrix, np.ndarray):
            raise TypeError(f"Expected state_matrix to be np.ndarray, got {type(state_matrix).__name__} instead.")

        differences = np.diff(state_matrix, axis=1) != 0
        fitness_values = np.sum(differences, axis=1)
        return fitness_values

    def get_problem_type(self) -> str:
        """Return the problem type.

        Returns
        -------
        str
            Specifies problem type as 'discrete'.
        """
        return self.problem_type



"""Unit tests for algorithms/"""

# Authors: Genevieve Hayes (modified by Kyle Nakamura)
# License: BSD 3 clause

import numpy as np

try:
    import mlrose_ky
except ImportError:
    import sys

    sys.path.append("..")
    import mlrose_ky

from mlrose_ky import OneMax, DiscreteOpt, ContinuousOpt, hill_climb, random_hill_climb, simulated_annealing, genetic_alg, mimic

SEED = 12


def test_mimic_discrete_max():
    """Test mimic function for a discrete maximization problem"""
    problem = DiscreteOpt(5, OneMax())
    best_state, best_fitness, _ = mimic(problem, random_state=SEED)
    x = np.ones(5)
    assert np.array_equal(best_state, x) and best_fitness == 5


def test_mimic_discrete_min():
    """Test mimic function for a discrete minimization problem"""
    problem = DiscreteOpt(5, OneMax(), maximize=False)
    best_state, best_fitness, _ = mimic(problem, random_state=SEED)
    x = np.zeros(5)
    assert np.array_equal(best_state, x) and best_fitness == 0


def test_hill_climb_discrete_max():
    """Test hill_climb function for a discrete maximization problem"""
    problem = DiscreteOpt(5, OneMax())
    best_state, best_fitness, _ = hill_climb(problem, restarts=10, random_state=SEED)
    x = np.ones(5)
    assert np.array_equal(best_state, x) and best_fitness == 5


def test_hill_climb_continuous_max():
    """Test hill_climb function for a continuous maximization problem"""
    problem = ContinuousOpt(5, OneMax())
    best_state, best_fitness, _ = hill_climb(problem, restarts=10, random_state=SEED)
    x = np.ones(5)
    assert np.array_equal(best_state, x) and best_fitness == 5


def test_hill_climb_discrete_min():
    """Test hill_climb function for a discrete minimization problem"""
    problem = DiscreteOpt(5, OneMax(), maximize=False)
    best_state, best_fitness, _ = hill_climb(problem, restarts=10, random_state=SEED)
    x = np.zeros(5)
    assert np.array_equal(best_state, x) and best_fitness == 0


def test_hill_climb_continuous_min():
    """Test hill_climb function for a continuous minimization problem"""
    problem = ContinuousOpt(5, OneMax(), maximize=False)
    best_state, best_fitness, _ = hill_climb(problem, restarts=10, random_state=SEED)
    x = np.zeros(5)
    assert np.array_equal(best_state, x) and best_fitness == 0


def test_hill_climb_max_iters():
    """Test hill_climb function with max_iters less than infinite"""
    problem = DiscreteOpt(5, OneMax())
    x = np.zeros(5)
    best_state, best_fitness, _ = hill_climb(problem, max_iters=1, init_state=x, random_state=SEED)
    assert best_fitness == 1


def test_random_hill_climb_discrete_max():
    """Test random_hill_climb function for a discrete maximization problem"""
    problem = DiscreteOpt(5, OneMax())
    best_state, best_fitness, _ = random_hill_climb(problem, restarts=10, random_state=SEED)
    x = np.ones(5)
    assert np.array_equal(best_state, x) and best_fitness == 5


def test_random_hill_climb_continuous_max():
    """Test random_hill_climb function for a continuous maximization problem"""
    problem = ContinuousOpt(5, OneMax())
    best_state, best_fitness, _ = random_hill_climb(problem, restarts=10, random_state=SEED)
    x = np.ones(5)
    assert np.array_equal(best_state, x) and best_fitness == 5


def test_random_hill_climb_discrete_min():
    """Test random_hill_climb function for a discrete minimization problem"""
    problem = DiscreteOpt(5, OneMax(), maximize=False)
    best_state, best_fitness, _ = random_hill_climb(problem, restarts=10, random_state=SEED)
    x = np.zeros(5)
    assert np.array_equal(best_state, x) and best_fitness == 0


def test_random_hill_climb_continuous_min():
    """Test random_hill_climb function for a continuous minimization problem"""
    problem = ContinuousOpt(5, OneMax(), maximize=False)
    best_state, best_fitness, _ = random_hill_climb(problem, restarts=10, random_state=SEED)
    x = np.zeros(5)
    assert np.array_equal(best_state, x) and best_fitness == 0


def test_random_hill_climb_max_iters():
    """Test random_hill_climb function with low max_iters"""
    problem = DiscreteOpt(5, OneMax())
    x = np.zeros(5)
    best_state, best_fitness, _ = random_hill_climb(problem, max_attempts=1, max_iters=1, init_state=x, random_state=SEED)
    assert best_fitness == 1


def test_simulated_annealing_discrete_max():
    """Test simulated_annealing function for a discrete maximization problem"""
    problem = DiscreteOpt(5, OneMax())
    best_state, best_fitness, _ = simulated_annealing(problem, random_state=SEED)
    x = np.ones(5)
    assert np.array_equal(best_state, x) and best_fitness == 5


def test_simulated_annealing_continuous_max():
    """Test simulated_annealing function for a continuous maximization problem"""
    problem = ContinuousOpt(5, OneMax())
    best_state, best_fitness, _ = simulated_annealing(problem, max_attempts=20, random_state=SEED)
    x = np.ones(5)
    assert np.array_equal(best_state, x) and best_fitness == 5


def test_simulated_annealing_discrete_min():
    """Test simulated_annealing function for a discrete minimization problem"""
    problem = DiscreteOpt(5, OneMax(), maximize=False)
    best_state, best_fitness, _ = simulated_annealing(problem, random_state=SEED)
    x = np.zeros(5)
    assert np.array_equal(best_state, x) and best_fitness == 0


def test_simulated_annealing_continuous_min():
    """Test simulated_annealing function for a continuous minimization problem"""
    problem = ContinuousOpt(5, OneMax(), maximize=False)
    best_state, best_fitness, _ = simulated_annealing(problem, random_state=SEED)
    x = np.zeros(5)
    assert np.array_equal(best_state, x) and best_fitness == 0


def test_simulated_annealing_max_iters():
    """Test simulated_annealing function with low max_iters"""
    problem = DiscreteOpt(5, OneMax())
    x = np.zeros(5)
    best_state, best_fitness, _ = simulated_annealing(problem, max_attempts=1, max_iters=1, init_state=x, random_state=SEED)
    assert best_fitness == 1


def test_genetic_alg_discrete_max():
    """Test genetic_alg function for a discrete maximization problem"""
    problem = DiscreteOpt(5, OneMax())
    best_state, best_fitness, _ = genetic_alg(problem, random_state=SEED)
    x = np.ones(5)
    assert np.array_equal(best_state, x) and best_fitness == 5


def test_genetic_alg_continuous_max():
    """Test genetic_alg function for a continuous maximization problem"""
    problem = ContinuousOpt(5, OneMax())
    best_state, best_fitness, _ = genetic_alg(problem, random_state=SEED)
    x = np.ones(5)
    assert np.allclose(best_state, x, atol=0.5) and best_fitness > 4


def test_genetic_alg_discrete_min():
    """Test genetic_alg function for a discrete minimization problem"""
    problem = DiscreteOpt(5, OneMax(), maximize=False)
    best_state, best_fitness, _ = genetic_alg(problem, random_state=SEED)
    x = np.zeros(5)
    assert np.array_equal(best_state, x) and best_fitness == 0


def test_genetic_alg_continuous_min():
    """Test genetic_alg function for a continuous minimization problem"""
    problem = ContinuousOpt(5, OneMax(), maximize=False)
    best_state, best_fitness, _ = genetic_alg(problem, random_state=SEED)
    x = np.zeros(5)
    assert np.allclose(best_state, x, atol=0.5) and best_fitness < 1



"""Functions to implement the randomized optimization and search algorithms."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

import numpy as np
from typing import Callable, Any
from mlrose_ky.decorators import short_name
from mlrose_ky.algorithms.decay import GeometricDecay


@short_name("sa")
def simulated_annealing(
    problem: Any,
    schedule: Any = GeometricDecay(),
    max_attempts: int = 10,
    max_iters: int = np.inf,
    init_state: np.ndarray = None,
    curve: bool = False,
    random_state: int = None,
    state_fitness_callback: Callable = None,
    callback_user_info: Any = None,
) -> tuple[np.ndarray, float, np.ndarray | None]:
    """Use simulated annealing to find the optimum for a given optimization problem.

    Parameters
    ----------
    problem: optimization object
        Object containing fitness function optimization problem to be solved.
        For example, :code:`DiscreteOpt()`, :code:`ContinuousOpt()` or
        :code:`TSPOpt()`.
    schedule: schedule object, default: :code:`mlrose_ky.GeometricDecay()`
        Schedule used to determine the value of the temperature parameter.
    max_attempts: int, default: 10
        Maximum number of attempts to find a better neighbor at each step.
    max_iters: int, default: np.inf
        Maximum number of iterations of the algorithm.
    init_state: np.ndarray, default: None
        1-D Numpy array containing starting state for algorithm.
        If :code:`None`, then a random state is used.
    curve: bool, default: False
        Boolean to keep fitness values for a curve.
        If :code:`False`, then no curve is stored.
        If :code:`True`, then a history of fitness values is provided as a
        third return value.
    random_state: int, default: None
        If random_state is a positive integer, random_state is the seed used
        by np.random.seed(); otherwise, the random seed is not set.
    state_fitness_callback: function taking five parameters, default: None
        If specified, this callback will be invoked once per iteration.
        Parameters are (iteration, max attempts reached?, current best state, current best fit, user callback data).
        Return true to continue iterating, or false to stop.
    callback_user_info: any, default: None
        User data passed as last parameter of callback.

    Returns
    -------
    best_state: np.ndarray
        Numpy array containing state that optimizes the fitness function.
    best_fitness: float
        Value of fitness function at best state.
    fitness_curve: np.ndarray
        Numpy array containing the fitness at every iteration.
        Only returned if input argument :code:`curve` is :code:`True`.

    References
    ----------
    Russell, S. and P. Norvig (2010). *Artificial Intelligence: A Modern Approach*, 3rd edition.
    Prentice Hall, New Jersey, USA.
    """
    if not isinstance(max_attempts, int) or max_attempts < 0:
        raise ValueError(f"max_attempts must be a positive integer. Got {max_attempts}")
    if not (isinstance(max_iters, int) or max_iters == np.inf) or max_iters < 0:
        raise ValueError(f"max_iters must be a positive integer or np.inf. Got {max_iters}")
    if init_state is not None and len(init_state) != problem.get_length():
        raise ValueError(
            f"init_state must have the same length as the problem. Expected length {problem.get_length()}, got {len(init_state)}"
        )

    # Set random seed
    if isinstance(random_state, int) and random_state > 0:
        np.random.seed(random_state)

    # Initialize problem
    fitness_curve = []
    if init_state is None:
        problem.reset()
    else:
        problem.set_state(init_state)
    if state_fitness_callback is not None:
        # initial call with base data
        state_fitness_callback(
            iteration=0,
            state=problem.get_state(),
            fitness=problem.get_adjusted_fitness(),
            fitness_evaluations=problem.fitness_evaluations,
            user_data=callback_user_info,
        )

    attempts = 0
    iters = 0
    continue_iterating = True
    while attempts < max_attempts and iters < max_iters:
        temp = schedule.evaluate(iters)
        iters += 1
        problem.current_iteration += 1

        if temp == 0:
            break
        else:
            # Find random neighbor and evaluate fitness
            next_state = problem.random_neighbor()
            next_fitness = problem.eval_fitness(next_state)

            # Calculate delta E and change prob
            current_fitness = problem.get_fitness()
            delta_e = next_fitness - current_fitness
            prob = np.exp(delta_e / temp)

            # If best neighbor is an improvement or random value is less
            # than prob, move to that state and reset attempts counter
            if delta_e > 0 or np.random.uniform() < prob:
                problem.set_state(next_state)
                attempts = 0
            else:
                attempts += 1

        if curve:
            fitness_curve.append((problem.get_adjusted_fitness(), problem.fitness_evaluations))

        # invoke callback
        if state_fitness_callback is not None:
            max_attempts_reached = attempts == max_attempts or iters == max_iters or problem.can_stop()
            continue_iterating = state_fitness_callback(
                iteration=iters,
                attempt=attempts + 1,
                done=max_attempts_reached,
                state=problem.get_state(),
                fitness=problem.get_adjusted_fitness(),
                fitness_evaluations=problem.fitness_evaluations,
                curve=np.asarray(fitness_curve) if curve else None,
                user_data=callback_user_info,
            )

        # break out if requested
        if not continue_iterating:
            break

    best_fitness = problem.get_maximize() * problem.get_fitness()
    best_state = problem.get_state()

    if curve:
        return best_state, best_fitness, np.asarray(fitness_curve)

    return best_state, best_fitness, None



"""Classes for defining optimization problem objects."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

import numpy as np

from mlrose_ky.algorithms.crossovers import TSPCrossover
from mlrose_ky.algorithms.mutators import GeneSwapMutator
from mlrose_ky.fitness import TravellingSalesperson
from mlrose_ky.opt_probs.discrete_opt import DiscreteOpt


class TSPOpt(DiscreteOpt):
    """Class for defining travelling salesperson optimisation problems.

    Parameters
    ----------
    length: int
        Number of elements in state vector. Must equal number of nodes in the
        tour.

    fitness_fn: fitness function object, default: None
        Object to implement fitness function for optimization. If :code:`None`,
        then :code:`TravellingSales(coords=coords, distances=distances)` is
        used by default.

    maximize: bool, default: False
        Whether to maximize the fitness function.
        Set :code:`False` for minimization problem.

    coords: list of pairs, default: None
        Ordered list of the (x, y) co-ordinates of all nodes. This assumes
        that travel between all pairs of nodes is possible. If this is not the
        case, then use distances instead. This argument is ignored if
        fitness_fn is not :code:`None`.

    distances: list of triples, default: None
        List giving the distances, d, between all pairs of nodes, u and v, for
        which travel is possible, with each list item in the form (u, v, d).
        Order of the nodes does not matter, so (u, v, d) and (v, u, d) are
        considered to be the same. If a pair is missing from the list, it is
        assumed that travel between the two nodes is not possible. This
        argument is ignored if fitness_fn or coords is not :code:`None`.

    source_graph: networkx.Graph, default: None
    """

    def __init__(self, length=None, fitness_fn=None, maximize=False, coords: list = None, distances: list[tuple] = None, source_graph=None):
        if (fitness_fn is None) and (coords is None) and (distances is None):
            raise Exception("""At least one of fitness_fn, coords and""" + """ distances must be specified.""")
        elif fitness_fn is None:
            fitness_fn = TravellingSalesperson(coords=coords, distances=distances)
        self.distances: list[tuple] = distances
        self.coords: list | None = coords
        if length is None:
            if coords is not None:
                length = len(coords)
            elif distances is not None:
                length = len(set([x for (x, _, _) in distances] + [x for (_, x, _) in distances]))
        self.length: int = length
        DiscreteOpt.__init__(
            self, length, fitness_fn, maximize, max_val=length, crossover=TSPCrossover(self), mutator=GeneSwapMutator(self)
        )

        if self.fitness_fn.get_problem_type() != "tsp":
            raise Exception("""fitness_fn must have problem type 'tsp'.""")

        self.source_graph = source_graph
        self.prob_type = "tsp"

    @staticmethod
    def adjust_probs(probs):
        """Normalize a vector of probabilities so that the vector sums to 1.

        Parameters
        ----------
        probs: np.ndarray
            Vector of probabilities that may or may not sum to 1.

        Returns
        -------
        adj_probs: np.ndarray
            Vector of probabilities that sums to 1. Returns a zero vector if
            sum(probs) = 0.
        """
        sp = np.sum(probs)
        return np.zeros(np.shape(probs)) if sp == 0 else probs / sp

    def find_neighbors(self):
        """Find all neighbors of the current state."""
        self.neighbors = []

        for node1 in range(self.length - 1):
            for node2 in range(node1 + 1, self.length):
                neighbor = np.copy(self.state)

                neighbor[node1] = self.state[node2]
                neighbor[node2] = self.state[node1]
                self.neighbors.append(neighbor)

    def random(self):
        """Return a random state vector.

        Returns
        -------
        state: np.ndarray
            Randomly generated state vector.
        """
        state = np.random.permutation(self.length)

        return state

    def random_mimic(self):
        """Generate single MIMIC sample from probability density.

        Returns
        -------
        state: np.ndarray
            State vector of MIMIC random sample.
        """
        remaining = list(np.arange(self.length))
        state = np.zeros(self.length, dtype=np.int8)
        node_probs = np.copy(self.node_probs)

        # Get value of first element in new sample
        state[0] = np.random.choice(self.length, p=node_probs[0, 0])
        remaining.remove(state[0])
        node_probs[:, :, state[0]] = 0

        # Get sample order
        self.find_sample_order()
        sample_order = self.sample_order[1:]

        # Set values of remaining elements of state
        for i in sample_order:
            par_ind = self.parent_nodes[i - 1]
            par_value = state[par_ind]
            probs = node_probs[i, par_value]

            if np.sum(probs) == 0:
                next_node = np.random.choice(remaining)

            else:
                adj_probs = self.adjust_probs(probs)
                next_node = np.random.choice(self.length, p=adj_probs)

            state[i] = next_node
            remaining.remove(next_node)
            node_probs[:, :, next_node] = 0

        return state

    def random_neighbor(self):
        """Return random neighbor of current state vector.

        Returns
        -------
        neighbor: np.ndarray
            State vector of random neighbor.
        """
        neighbor = np.copy(self.state)
        node1, node2 = np.random.choice(np.arange(self.length), size=2, replace=False)

        neighbor[node1] = self.state[node2]
        neighbor[node2] = self.state[node1]

        return neighbor

    def sample_pop(self, sample_size):
        """Generate new sample from probability density.

        Parameters
        ----------
        sample_size: int
            Size of sample to be generated.

        Returns
        -------
        new_sample: np.ndarray
            Numpy array containing new sample.
        """
        if sample_size <= 0:
            raise Exception("""sample_size must be a positive integer.""")
        elif not isinstance(sample_size, int):
            if sample_size.is_integer():
                sample_size = int(sample_size)
            else:
                raise Exception("""sample_size must be a positive integer.""")

        self.find_sample_order()
        new_sample = []

        for _ in range(sample_size):
            state = self.random_mimic()
            new_sample.append(state)

        new_sample = np.array(new_sample)

        return new_sample



import mlrose_ky
from mlrose_ky.decorators import short_name
from mlrose_ky.runners._runner_base import _RunnerBase

"""
Example usage:

    experiment_name = 'example_experiment'
    problem = TSPGenerator.generate(seed=SEED, number_of_cities=22)

    ga = GARunner(problem=problem,
                  experiment_name=experiment_name,
                  output_directory=OUTPUT_DIRECTORY,
                  seed=SEED,
                  iteration_list=2 ** np.arange(12),
                  max_attempts=1000,
                  population_sizes=[150, 200, 300],
                  mutation_rates=[0.4, 0.5, 0.6])
                  
    # the two data frames will contain the results
    df_run_stats, df_run_curves = ga.run()                  
"""


@short_name("ga")
class GARunner(_RunnerBase):

    def __init__(
        self,
        problem,
        experiment_name,
        seed,
        iteration_list,
        population_sizes,
        mutation_rates,
        hamming_factors=None,
        hamming_factor_decays=None,
        max_attempts=500,
        generate_curves=True,
        **kwargs,
    ):
        super().__init__(
            problem=problem,
            experiment_name=experiment_name,
            seed=seed,
            iteration_list=iteration_list,
            max_attempts=max_attempts,
            generate_curves=generate_curves,
            **kwargs,
        )
        self.population_sizes = population_sizes
        self.mutation_rates = mutation_rates
        self.hamming_factors = hamming_factors
        self.hamming_factor_decays = hamming_factor_decays

    def run(self):
        return super().run_experiment_(
            algorithm=mlrose_ky.genetic_alg,
            pop_size=("Population Size", self.population_sizes),
            mutation_prob=("Mutation Rate", self.mutation_rates),
            hamming_factor=("Hamming Factor", self.hamming_factors),
            hamming_decay_factor=("Hamming Factor Decay Rate", self.hamming_factor_decays),
        )



"""Classes for defining neural network weight optimization problems."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause
from abc import ABC, abstractmethod

import numpy as np
from sklearn.base import BaseEstimator

from mlrose_ky.neural.fitness.network_weights import NetworkWeights
from mlrose_ky.neural.utils import unflatten_weights
from mlrose_ky.opt_probs import ContinuousOpt


class _NNBase(BaseEstimator, ABC):

    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def fit(self, X, y=None, init_weights=None):
        """Fit neural network to data.

        Parameters
        ----------
        X: np.ndarray
            Numpy array containing feature dataset with each row
            representing a single observation.

        y: np.ndarray
            Numpy array containing data labels. Length must be same as
            length of X.

        init_weights: np.ndarray, default: None
            Numpy array containing starting weights for algorithm.
            If :code:`None`, then a random state is used.
        """
        pass

    @abstractmethod
    def predict(self, X):
        """Use model to predict data labels for given feature array.

        Parameters
        ----------
        X: np.ndarray
            Numpy array containing feature dataset with each row
            representing a single observation.

        Returns
        -------
        y_pred: np.ndarray
            Numpy array containing predicted data labels.
        """
        pass

    @staticmethod
    def _calculate_state_size(node_list):
        num_nodes = 0
        for i in range(len(node_list) - 1):
            num_nodes += node_list[i] * node_list[i + 1]
        return num_nodes

    @staticmethod
    def _build_node_list(X, y, hidden_nodes, bias):
        # Determine number of nodes in each layer
        input_nodes = np.shape(X)[1] + bias
        output_nodes = np.shape(y)[1]
        node_list = [input_nodes] + list(hidden_nodes) + [output_nodes]
        return node_list

    @staticmethod
    def _format_x_y_data(X, y):
        # Make sure y is an array and not a list
        y = np.array(y)
        # Convert y to 2D if necessary
        if len(np.shape(y)) == 1:
            y = np.reshape(y, [len(y), 1])
        # Verify X and y are the same length
        if not np.shape(X)[0] == np.shape(y)[0]:
            raise Exception("The length of X and y must be equal.")
        return X, y

    @staticmethod
    def _build_problem_and_fitness_function(X, y, node_list, activation, learning_rate, bias, clip_max, is_classifier=True):
        # Initialize optimization problem
        fitness = NetworkWeights(X, y, node_list, activation, bias, is_classifier, learning_rate=learning_rate)
        num_nodes = _NNBase._calculate_state_size(node_list)
        problem = ContinuousOpt(
            length=num_nodes, fitness_fn=fitness, maximize=False, min_val=-1 * clip_max, max_val=clip_max, step=learning_rate
        )
        return fitness, problem

    @staticmethod
    def _predict(X, fitted_weights, node_list, bias, input_activation, output_activation, is_classifier=True):
        weights = list(unflatten_weights(fitted_weights, node_list))

        # Add bias column to inputs matrix, if required
        if bias:
            ones = np.ones([np.shape(X)[0], 1])
            inputs = np.hstack((X, ones))

        else:
            inputs = X

        # Pass data through network
        y_pred = None
        for i in range(len(weights)):
            # Multiple inputs by weights
            outputs = np.dot(inputs, weights[i])

            # Transform outputs to get inputs for next layer (or final preds)
            if i < len(weights) - 1:
                inputs = input_activation(outputs)
            else:
                y_pred = output_activation(outputs)

        # For classifier, convert predicted probabilities to 0-1 labels
        predicted_probs = None
        if is_classifier:
            predicted_probs = y_pred

            if node_list[-1] == 1:
                y_pred = np.round(y_pred).astype(int)
            else:
                zeros = np.zeros_like(y_pred)
                zeros[np.arange(len(y_pred)), np.argmax(y_pred, axis=1)] = 1
                y_pred = zeros.astype(int)

        return y_pred, predicted_probs



from abc import ABC, abstractmethod
import time
import os
import logging
import itertools as it
from typing import Any

import numpy as np
import pandas as pd
import pickle as pk
import inspect as lk
import signal
import multiprocessing
import ctypes

from mlrose_ky.decorators import get_short_name
from mlrose_ky.runners.utils import build_data_filename


class _RunnerBase(ABC):
    __abort = multiprocessing.Value(ctypes.c_bool)
    __spawn_count = multiprocessing.Value(ctypes.c_uint)
    __replay = multiprocessing.Value(ctypes.c_bool)
    __original_sigint_handler = None
    __sigint_params = None

    @classmethod
    def runner_name(cls):
        return get_short_name(cls)

    def dynamic_runner_name(self):
        return self.__dynamic_short_name__ if hasattr(self, "__dynamic_short_name__") else self.runner_name()

    def _set_dynamic_runner_name(self, name):
        self.__dynamic_short_name__ = name

    @staticmethod
    def _print_banner(text):
        logging.info("*" * len(text))
        logging.info(text)
        logging.info("*" * len(text))

    @staticmethod
    def _sanitize_value(value):
        if isinstance(value, tuple) or isinstance(value, list):
            v = str(value)
        elif isinstance(value, np.ndarray):
            v = str(list(value))
        else:
            v = get_short_name(value)
        return v

    @abstractmethod
    def run(self):
        pass

    def __init__(
        self,
        problem,
        experiment_name,
        seed,
        iteration_list,
        max_attempts=500,
        generate_curves=True,
        output_directory=None,
        copy_zero_curve_fitness_from_first=False,
        replay=False,
        override_ctrl_c_handler=True,
        **kwargs,
    ):
        self.problem = problem
        self.seed = seed
        self.iteration_list = iteration_list
        self.max_attempts = max_attempts
        self.generate_curves = generate_curves
        self.parameter_description_dict = {}
        self.override_ctrl_c_handler = override_ctrl_c_handler

        self.run_stats_df = None
        self.curves_df = None
        self._raw_run_stats = []
        self._fitness_curves = []
        self._curve_base = 0
        self._copy_zero_curve_fitness_from_first = copy_zero_curve_fitness_from_first
        self._copy_zero_curve_fitness_from_first_original = copy_zero_curve_fitness_from_first
        self._extra_args = kwargs
        self._output_directory = output_directory
        self._experiment_name = experiment_name
        self._current_logged_algorithm_args = {}
        self._run_start_time = None
        self._iteration_times = []
        self._first_curve_synthesized = False
        if replay:
            self.set_replay_mode()
        self._increment_spawn_count()

    def _increment_spawn_count(self):
        with self.__spawn_count.get_lock():
            self.__spawn_count.value += 1

    def _decrement_spawn_count(self):
        with self.__spawn_count.get_lock():
            self.__spawn_count.value -= 1

    def _get_spawn_count(self):
        self._print_banner(f"*** Spawn Count Remaining: {self.__spawn_count.value} ***")
        return self.__spawn_count.value

    def abort(self):
        self._print_banner("*** ABORTING ***")
        with self.__abort.get_lock():
            self.__abort.value = True

    def has_aborted(self):
        return self.__abort.value

    def set_replay_mode(self, value=True):
        with self.__replay.get_lock():
            self.__replay.value = value

    def replay_mode(self):
        return self.__replay.value

    def _setup(self):
        self._raw_run_stats = []
        self._fitness_curves = []
        self._curve_base = 0

        self._iteration_times = []
        self._copy_zero_curve_fitness_from_first = self._copy_zero_curve_fitness_from_first_original
        self._current_logged_algorithm_args.clear()
        if self._output_directory is not None:
            if not os.path.exists(self._output_directory):
                os.makedirs(self._output_directory)

        # set up ctrl-c handler
        if self.override_ctrl_c_handler:
            if self.__original_sigint_handler is None:
                self.__original_sigint_handler = signal.getsignal(signal.SIGINT)
                signal.signal(signal.SIGINT, self._ctrl_c_handler)

    def _ctrl_c_handler(self, sig, frame):
        logging.info("Interrupted - saving progress so far")
        self.__sigint_params = (sig, frame)
        self.abort()

    def _tear_down(self):
        if not self.override_ctrl_c_handler:
            return
        try:
            # restore ctrl-c handler
            self._decrement_spawn_count()
            if self.__original_sigint_handler is not None:
                signal.signal(signal.SIGINT, self.__original_sigint_handler)
                if self.has_aborted() and self._get_spawn_count() == 0:
                    sig, frame = self.__sigint_params
                    self.__original_sigint_handler(sig, frame)
        except (ValueError, TypeError, AttributeError, Exception) as e:
            logging.error(f"Problem restoring SIGNINT handler: {e}")

    def log_current_argument(self, arg_name, arg_value):
        self._current_logged_algorithm_args[arg_name] = arg_value

    def run_experiment_(self, algorithm, **kwargs):
        self._setup()
        # extract loop params
        values = [([(k, v) for v in vs]) for (k, (n, vs)) in kwargs.items() if vs is not None]
        self.parameter_description_dict = {k: n for (k, (n, vs)) in kwargs.items() if vs is not None}
        value_sets = list(it.product(*values))

        logging.info(f"Running {self.dynamic_runner_name()}")
        run_start = time.perf_counter()
        for vns in value_sets:
            total_args = dict(vns)
            if "max_iters" not in total_args:
                total_args["max_iters"] = int(max(self.iteration_list))

            self._run_one_experiment(algorithm, total_args)

        run_end = time.perf_counter()
        logging.info(f"Run time: {run_end - run_start}")

        self._create_and_save_run_data_frames(final_save=True)
        self._tear_down()

        return self.run_stats_df, self.curves_df

    def _run_one_experiment(self, algorithm, total_args, **params):
        if self._extra_args is not None and len(self._extra_args) > 0:
            total_args.update(self._extra_args)
        total_args.update(params)

        user_info = [(k, v) for k, v in total_args.items()]
        self._invoke_algorithm(
            algorithm=algorithm,
            problem=self.problem,
            max_attempts=self.max_attempts,
            curve=self.generate_curves,
            user_info=user_info,
            **total_args,
        )

    def _create_and_save_run_data_frames(self, extra_data_frames=None, final_save=False):
        self.run_stats_df = pd.DataFrame(self._raw_run_stats)
        self.curves_df = pd.DataFrame(self._fitness_curves)
        if self._output_directory is not None:
            if len(self.run_stats_df) > 0:
                self._dump_df_to_disk(self.run_stats_df, df_name="run_stats_df", final_save=final_save)
            if self.generate_curves and len(self.curves_df) > 0:
                self._dump_df_to_disk(self.curves_df, df_name="curves_df", final_save=final_save)
            # output any extra
            if isinstance(extra_data_frames, dict):
                for n, v in extra_data_frames.items():
                    self._dump_df_to_disk(v, df_name=n, final_save=final_save)

    def _dump_df_to_disk(self, df, df_name, final_save=False):
        filename_root = self._dump_pickle_to_disk(object_to_pickle=df, name=df_name)
        df.to_csv(f"{filename_root}.csv")
        if final_save:
            logging.info(f"Saving: [{filename_root}.csv]")

    def _get_pickle_filename_root(self, name):
        filename_root = build_data_filename(
            output_directory=self._output_directory,
            runner_name=self.dynamic_runner_name(),
            experiment_name=self._experiment_name,
            df_name=name,
        )
        return filename_root

    def _dump_pickle_to_disk(self, object_to_pickle, name, final_save=False):
        if self._output_directory is None:
            return
        filename_root = self._get_pickle_filename_root(name)

        pk.dump(object_to_pickle, open(f"{filename_root}.p", "wb"))
        if final_save:
            logging.info(f"Saving: [{filename_root}.p]")
        return filename_root

    def _load_pickles(self):
        curves_df_filename = f"{self._get_pickle_filename_root('curves_df')}.p"
        run_stats_df_filename = f"{self._get_pickle_filename_root('run_stats_df')}.p"

        self.curves_df = None
        self.run_stats_df = None

        if os.path.exists(curves_df_filename):
            with open(curves_df_filename, "rb") as pickle_file:
                try:
                    self.curves_df = pk.load(pickle_file)
                except (OSError, IOError, pk.PickleError):
                    pass

        if os.path.exists(run_stats_df_filename):
            with open(run_stats_df_filename, "rb") as pickle_file:
                try:
                    self.run_stats_df = pk.load(pickle_file)
                except (OSError, IOError, pk.PickleError):
                    pass

        return self.curves_df is not None and self.run_stats_df is not None

    def _invoke_algorithm(self, algorithm, problem, max_attempts, curve, user_info, additional_algorithm_args=None, **total_args):
        self._current_logged_algorithm_args.update(total_args)
        if additional_algorithm_args is not None:
            self._current_logged_algorithm_args.update(additional_algorithm_args)

        if self.replay_mode() and self._load_pickles():
            return None, None, None

        self._print_banner("*** Run START ***")
        np.random.seed(self.seed)

        valid_args = [k for k in lk.signature(algorithm).parameters]
        args_to_pass = {k: v for k, v in total_args.items() if k in valid_args}

        self.start_run_timing()
        problem.reset()
        ret = algorithm(
            problem=problem,
            max_attempts=max_attempts,
            curve=curve,
            random_state=self.seed,
            state_fitness_callback=self.save_state,
            callback_user_info=user_info,
            **args_to_pass,
        )

        self._print_banner("*** Run END ***")
        self._curve_base = len(self._fitness_curves)

        return ret

    def start_run_timing(self):
        self._run_start_time = time.perf_counter()

    @staticmethod
    def _create_curve_stat(iteration, curve_value, curve_data, t=None):
        curve_fitness_value, curve_feval_value = curve_value
        curve_stat = {"Iteration": iteration, "Time": t, "Fitness": curve_fitness_value, "FEvals": curve_feval_value}

        curve_stat.update(curve_data)
        if isinstance(curve_value, dict):
            curve_stat.update(curve_value)

        return curve_stat

    def save_state(self, iteration, state, fitness, user_data, attempt=0, done=False, curve=None, fitness_evaluations=None):

        # log iteration timing
        end = time.perf_counter()
        t = end - self._run_start_time
        self._iteration_times.append(t)

        # FIXME: do we need to log anything else?
        if iteration > 0 and iteration not in self.iteration_list and not done:
            return True

        display_data = {**self._current_logged_algorithm_args}
        if user_data is not None and len(user_data) > 0:
            display_data.update({n: v for (n, v) in user_data})
            data_desc = ", ".join([f"{n}:[{get_short_name(v)}]" for n, v in display_data.items()])
            logging.debug(data_desc)
        logging.debug(
            f"runner_name:[{self.dynamic_runner_name()}], experiment_name:[{self._experiment_name}], "
            + ("" if attempt is None else f"attempt:[{attempt}], ")
            + f"iteration:[{iteration}], done:[{done}], "
            f"time:[{t:.2f}], fitness:[{fitness:.4f}]"
        )

        state_string = str(state).replace("\n", "//")[:200]
        logging.debug(f"\t{state_string}...")
        logging.debug("")

        def gd(n):
            return n if n not in self.parameter_description_dict.keys() else self.parameter_description_dict[n]

        def gi(v: Any):
            return v.get_info__(t) if hasattr(v, "get_info__") else {}

        current_iteration_stats = {str(gd(k)): self._sanitize_value(v) for k, v in self._current_logged_algorithm_args.items()}
        current_iteration_stats.update({str(gd(k)): self._sanitize_value(v) for k, v in {k: v for (k, v) in user_data}.items()})

        ai = (gi(v) for k, v in current_iteration_stats.items())
        additional_info = {k: self._sanitize_value(v) for d in ai for k, v in d.items()}

        if iteration > 0:
            remaining_iterations = [i for i in self.iteration_list if i >= iteration]
            iterations = [min(remaining_iterations)] if not done else remaining_iterations
        else:
            iterations = [0]

        for i in iterations:
            run_stat = {"Iteration": i, "Fitness": fitness, "FEvals": fitness_evaluations, "Time": t, "State": self._sanitize_value(state)}
            run_stat.update(additional_info)
            run_stat.update(current_iteration_stats)
            self._raw_run_stats.append(run_stat)

        if self.generate_curves and iteration == 0:
            # capture first fitness value for iteration 0 if not already captured.
            if curve is None or len(curve) == 0:
                curve = [(fitness, fitness_evaluations)]
                self._first_curve_synthesized = True

        if self.generate_curves and curve is not None:  # and (done or iteration == max(self.iteration_list)):
            curve_stats_saved = len(self._fitness_curves)
            total_curve_stats = self._curve_base + len(curve)
            curve_stats_to_save = total_curve_stats - curve_stats_saved
            if self._first_curve_synthesized:
                curve_stats_to_save += 1
            ix_end = iteration + 1
            ix_start = ix_end - curve_stats_to_save
            if ix_start < 0:
                ix_start = 0

            fc = list(zip(range(ix_start, ix_end), curve[-curve_stats_to_save:]))

            curve_stats = [
                self._create_curve_stat(iteration=ix, curve_value=f, curve_data=current_iteration_stats, t=self._iteration_times[ix])
                for (ix, f) in fc
            ]

            self._fitness_curves.extend(curve_stats)

            if self._copy_zero_curve_fitness_from_first and len(self._fitness_curves) > 1:
                self._fitness_curves[0]["Fitness"] = self._fitness_curves[1]["Fitness"]
                self._copy_zero_curve_fitness_from_first = False
            self._create_and_save_run_data_frames()

        return not (self.has_aborted() or done)



"""Classes for defining mutation strategies for Genetic Algorithms (GA)."""

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

from .discrete_mutator import DiscreteGeneMutator
from .gene_swap_mutator import GeneSwapMutator
from .single_gene_mutator import SingleGeneMutator
from .single_shift_mutator import SingleShiftMutator
